import contextlib
import copy
import inspect
import io
import json
import math
import os
import re
import sqlite3
import stat
import sys
import tarfile
import threading
import time
import traceback
import urllib.parse
from timeit import default_timer as timer
from typing import IO, Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union, cast

with contextlib.suppress(ImportError):
    import rapidgzip

from ratarmountcore.BlockParallelReaders import ParallelXZReader
from ratarmountcore.compressions import COMPRESSION_BACKENDS, get_gzip_info, open_compressed_file
from ratarmountcore.formats import FileFormatID, might_be_format
from ratarmountcore.mountsource import FileInfo, MountSource
from ratarmountcore.mountsource.SQLiteIndexMountSource import SQLiteIndexMountSource
from ratarmountcore.ProgressBar import ProgressBar
from ratarmountcore.SQLiteIndex import SQLiteIndex, SQLiteIndexedTarUserData
from ratarmountcore.StenciledFile import RawStenciledFile, StenciledFile
from ratarmountcore.utils import (
    CompressionError,
    InvalidIndexError,
    RatarmountError,
    ceil_div,
    decode_unpadded_base64,
    determine_recursion_depth,
    get_xdg_cache_home,
    overrides,
)

BLOCK_SIZE = 512
# Note that the number of decimal digits for UINT64_MAX is ceil[64*ln(2)/ln(10)] = 20.
_PAX_HEADER_LENGTH_PREFIX_REGEX = re.compile(b"([0-9]{1,20}) ")
_PAX_HEADER_SIZE_KEYWORD = re.compile(br" size=([0-9]{1,20})\n")


# Patch https://github.com/python/cpython/issues/136602
def patch_tarfile():
    _proc_gnusparse_10 = getattr(tarfile.TarInfo, '_proc_gnusparse_10', None)
    if (
        not callable(_proc_gnusparse_10)
        or not hasattr(tarfile.TarInfo, '_link_target')
        or inspect.getfullargspec(_proc_gnusparse_10).args != ['self', 'next', 'pax_headers', 'tarfile']
    ):
        return

    # pylint: disable=redefined-builtin,redefined-outer-name,protected-access
    def _wrapped_proc_gnusparse_10(self, next, pax_headers, tarfile, *args, **kwargs):  # noqa
        # Abuse _link_target because I am too inept to dynamically extend TarInfo.__slots__.
        # This member is only used for extracting anyway and we only use extract* methods for sparse files.
        # And it makes wrapping __init__ unnecessary.
        next._link_target = next.offset_data
        return _proc_gnusparse_10(self, next, pax_headers, tarfile, *args, **kwargs)

    tarfile.TarInfo._proc_gnusparse_10 = _wrapped_proc_gnusparse_10


with contextlib.suppress(Exception):
    patch_tarfile()


class _TarFileMetadataReader:
    def __init__(
        self,
        parent: 'SQLiteIndexedTar',
        setFileInfos: Callable[[List[Tuple]], None],
        setxattrs: Callable[[List[Tuple]], None],
        updateProgressBar: Callable[[], None],
        recursionDepth: int,
    ):
        self._parent = parent
        self._set_file_infos = setFileInfos
        self._setxattrs = setxattrs
        self._update_progress_bar = updateProgressBar
        self._recursionDepth = recursionDepth
        self._lastUpdateTime = time.time()

    @staticmethod
    def _get_tar_prefix(fileObject: IO[bytes], tarInfo: tarfile.TarInfo, printDebug: int) -> Optional[bytes]:
        """Get the actual prefix as stored in the TAR."""

        # Offsets taken from https://en.wikipedia.org/wiki/Tar_(computing)#UStar_format
        def extract_prefix(tarBlockOffset):
            fileObject.seek(tarBlockOffset + 345)
            return fileObject.read(155)

        def extract_name(tarBlockOffset):
            fileObject.seek(tarBlockOffset)
            return fileObject.read(100)

        def extract_size(tarBlockOffset):
            fileObject.seek(tarBlockOffset + 124)
            return int(fileObject.read(12).strip(b"\0"), 8)  # octal encoded file size TODO might also be base64

        oldPosition = fileObject.tell()

        # Normally, getting the prefix, could be as easy as calling extract_prefix.
        # But, for long-names the prefix will not be prefixed but for long links it will be prefixed by tarfile.
        # This complicates things. Also, both long link and long name are implemented by a prepended
        # tar block with the special file name "././@LongLink" and tarfile will return the header offset of the
        # corresponding GNU LongLink file header in the TarInfo object instead of the actual file header, which
        # contains the prefix.
        try:
            if extract_name(tarInfo.offset).startswith(b"././@LongLink\0"):
                nextHeaderOffset = tarInfo.offset + 512 + (extract_size(tarInfo.offset) + 512 - 1) // 512 * 512
                return extract_prefix(nextHeaderOffset)
            return extract_prefix(tarInfo.offset)

        except Exception as exception:
            if printDebug >= 1:
                print("[Warning] Encountered exception when trying to get TAR prefix", exception)
            if printDebug >= 3:
                traceback.print_exc()

        finally:
            fileObject.seek(oldPosition)

        return None

    @staticmethod
    def _tar_info_full_mode(tarInfo: tarfile.TarInfo) -> int:
        """
        Returns the full mode for a TarInfo object. Note that TarInfo.mode only contains the permission bits
        and not other bits like set for directory, symbolic links, and other special files.
        """

        # fmt: off
        return (
            tarInfo.mode
            | ( stat.S_IFDIR if tarInfo.isdir () else 0 )
            | ( stat.S_IFREG if tarInfo.isfile() or tarInfo.type == b'D' else 0 )
            | ( stat.S_IFLNK if tarInfo.issym () else 0 )
            | ( stat.S_IFCHR if tarInfo.ischr () else 0 )
            | ( stat.S_IFIFO if tarInfo.isfifo() else 0 )
        )
        # fmt: on

    @staticmethod
    def _fix_incremental_backup_name_prefixes(fileObject: IO[bytes], tarInfo: tarfile.TarInfo, printDebug: int):
        """
        Tarfile joins the TAR prefix with the file path.
        However, for incremental TARs, the prefix is an octal timestamp and should be ignored.
        This function reads the raw prefix from the TAR file and removes it from the TarInfo object's path
        if the prefix is an octal number indicating an incremental archive prefix.
        """

        if '/' not in tarInfo.name:
            return

        fixedPath = None
        prefix, name = tarInfo.name.split('/', 1)

        realPrefix = _TarFileMetadataReader._get_tar_prefix(fileObject, tarInfo, printDebug)
        encodedPrefix = prefix.encode('utf8', 'surrogateescape')

        # For names longer than 100B, GNU tar will store it using a ././@LongLink named file.
        # In this case, tarfile will ignore the truncated filename AND the octal timestamp prefix!
        # However, for long symbolic links, the prefix WILL be prepended to the @LongLink contents!
        # In order to not strip folders erroneously, test against this prefix. Unfortunately, this is
        # not perfect either because tarfile removes trailing slashes for names. So we have to
        # read the TAR information ourselves.
        # Note that the prefix contains two not always identical octal timestamps! E.g.,
        #   b'13666753432\x0013666377326\x00\x00\x00...
        # We only test for the first here as I'm not sure what the second one is.
        # In some cases instead of the octal timestamp there will be unknown binary data!
        # Because of this the data is not asserted to be octal.
        if realPrefix and realPrefix.startswith(encodedPrefix + b"\0"):
            fixedPath = name

        if fixedPath is None and printDebug >= 1:
            print(f"[Warning] ignored prefix '{encodedPrefix!r}' because it was not found in TAR header prefix.")
            print("[Warning]", realPrefix[:30] if realPrefix else realPrefix)
            print(f"[Info] TAR header offset: {tarInfo.offset}, type: {tarInfo.type!s}")
            print("[Info] name:", tarInfo.name)
            print()

        if fixedPath is not None:
            tarInfo.name = fixedPath

    # fmt: off
    @staticmethod
    def _process_tar_info(
        tarInfo          : tarfile.TarInfo,
        fileObject       : IO[bytes],
        pathPrefix       : str,
        streamOffset     : int,
        isGnuIncremental : Optional[bool],
        mountRecursively : bool,
        transform        : Callable[[str], str],
        recursionDepth   : int,
        printDebug       : int,
    ) -> Tuple[List[Tuple], List[Tuple], bool, Optional[bool]]:
        # fmt: on
        """Postprocesses a TarInfo object into one or multiple FileInfo tuples."""

        if tarInfo.type == b'D' and not isGnuIncremental:
            isGnuIncremental = True
            if printDebug >= 1:
                print(f"[Warning] A folder metadata entry ({tarInfo.name}) for GNU incremental archives")
                print("[Warning] was encountered but this archive was not automatically recognized as such!")
                print("[Warning] Please call ratarmount with the --gnu-incremental flag if there are problems.")
                print()

        if isGnuIncremental:
            _TarFileMetadataReader._fix_incremental_backup_name_prefixes(fileObject, tarInfo, printDebug)

        offsetHeader = streamOffset + tarInfo.offset

        prefix = 'SCHILY.xattr.'
        xattrs = {
            key[len(prefix) :]: value.encode('utf-8', errors='surrogateescape')
            for key, value in tarInfo.pax_headers.items()
            if key.startswith(prefix)
        }
        prefix = 'LIBARCHIVE.xattr.'
        xattrs.update(
            {
                urllib.parse.unquote(key[len(prefix) :]): decode_unpadded_base64(value)
                for key, value in tarInfo.pax_headers.items()
                if key.startswith(prefix)
            }
        )
        xattrRows = [(offsetHeader, key, value) for key, value in xattrs.items()]

        fullPath = pathPrefix + "/" + tarInfo.name
        if tarInfo.isdir():
            fullPath += "/"
        path, name = SQLiteIndex.normpath(transform(fullPath)).rsplit("/", 1)

        # TODO: As for the tarfile type SQLite expects int but it is generally bytes.
        #       Most of them would be convertible to int like tarfile.SYMTYPE which is b'2',
        #       but others should throw errors, like GNUTYPE_SPARSE which is b'S'.
        #       When looking at the generated index, those values get silently converted to 0?
        # fmt: off
        fileInfo : Tuple = (
            path                                            ,  # 0  : path
            name                                            ,  # 1  : file name
            offsetHeader                                    ,  # 2  : header offset
            streamOffset + tarInfo.offset_data              ,  # 3  : data offset
            tarInfo.size                                    ,  # 4  : file size
            tarInfo.mtime                                   ,  # 5  : modification time
            _TarFileMetadataReader._tar_info_full_mode(tarInfo),  # 6  : file mode / permissions
            tarInfo.type                                    ,  # 7  : TAR file type. Currently unused.
            tarInfo.linkname                                ,  # 8  : linkname
            tarInfo.uid                                     ,  # 9  : user ID
            tarInfo.gid                                     ,  # 10 : group ID
            False                                           ,  # 11 : is TAR (unused?)
            tarInfo.issparse()                              ,  # 12 : is sparse
            False                                           ,  # 13 : is generated (parent folder)
            recursionDepth                                  ,  # 14 : recursion depth
        )
        # fmt: on

        fileInfos = [fileInfo]

        if mountRecursively and tarInfo.isfile() and tarInfo.name.lower().endswith('.tar'):
            return fileInfos, xattrRows, True, isGnuIncremental

        # Add GNU incremental TAR directory metadata files also as directories
        if tarInfo.type == b'D':
            dirFileInfo = list(fileInfo)
            # This is only to get a unique primary key :/
            # Then again, TAR blocks are known to be on 512B boundaries, so the lower
            # bits in the offset are redundant anyway.
            dirFileInfo[2] += 1
            dirFileInfo[4] = 0  # directory entries have no size by convention
            dirFileInfo[6] = tarInfo.mode | stat.S_IFDIR
            fileInfos.append(tuple(dirFileInfo))

        return fileInfos, xattrRows, False, isGnuIncremental

    @staticmethod
    def _get_pax_size(block: bytes) -> int:
        # See https://pubs.opengroup.org/onlinepubs/9699919799/utilities/pax.html#tag_20_92_13_03
        # > An extended header shall consist of one or more records, each constructed as follows:
        # >   "%d %s=%s\n", <length>, <keyword>, <value>
        # > The <length> field shall be the decimal length of the extended header record in octets,
        # > including the trailing <newline>.
        offset = 0
        pax_size = -1
        while offset < len(block) and block[offset] != 0:
            match_length = _PAX_HEADER_LENGTH_PREFIX_REGEX.match(block, offset)
            if not match_length:
                break

            length = int(match_length.group(1))  # Leading zeros are not a problem.
            # Shortest length is "5 x=\n"
            if length < 5 or offset + length > len(block):
                break

            match_size = _PAX_HEADER_SIZE_KEYWORD.match(block, match_length.end(1), offset + length)
            if match_size:
                pax_size = int(match_size.group(1))

            offset += length
        return pax_size

    @staticmethod
    def find_tar_file_offsets(fileObject: IO[bytes], ignoreZeros: bool) -> Generator[Tuple[int, bytes], None, None]:
        """
        Generator which yields offsets in the given TAR suitable for splitting the file into sub TARs.
        Also returns the type of the TAR metadata block at the returned offset for convenience.
        """

        block_number = 0
        skip_next_blocks = 0
        pax_size = -1
        fileObject.seek(0)

        while True:
            block_contents = fileObject.read(BLOCK_SIZE)
            block_number += 1
            if len(block_contents) < BLOCK_SIZE:
                break

            # > The end of an archive is marked by at least two consecutive zero-filled records.
            if not any(block_contents):
                block_contents = fileObject.read(BLOCK_SIZE)
                block_number += 1
                if not any(block_contents):
                    if ignoreZeros:
                        continue
                    break

                if len(block_contents) < BLOCK_SIZE:
                    break

            type_flag = block_contents[156:157]

            if skip_next_blocks > 0:
                skip_next_blocks -= 1
            else:
                yield (block_number - 1) * BLOCK_SIZE, type_flag

            raw_size = block_contents[124 : 124 + 12].strip(b"\0")
            size = int(raw_size, 8) if raw_size else 0
            if size == 0 and pax_size > 0:
                size = pax_size
            data_block_count = ceil_div(size, BLOCK_SIZE)

            pax_size = -1
            if type_flag in (b"x", b"g", b"X"):
                pax_size = _TarFileMetadataReader._get_pax_size(fileObject.read(data_block_count * BLOCK_SIZE))

            block_number += data_block_count
            fileObject.seek(block_number * BLOCK_SIZE)

            # A lot of the special files contain information about the next file, therefore keep do not yield
            # the offset of the next block so that the TAR will not be split between them.
            # K: Identifies the *next* file on the tape as having a long name.
            # L: Identifies the *next* file on the tape as having a long linkname.
            # x: Extended header with meta data for the next file in the archive (POSIX.1-2001)
            # X: Solaris extended header
            # g: Extended global header (POSIX.1-2001)
            # 0: Normal file.
            if type_flag != b'0':
                skip_next_blocks += 1

    def _open_tar(self, fileObject: IO[bytes]):
        """
        Opens the fileObject with the appropriate settings using the tarfile module.
        Instead of throwing, an empty iterable might be returned.
        """

        if not self._parent.isTar:
            return []  # Feign an empty TAR file (iterable) if anything goes wrong

        try:
            # r: uses seeks to skip to the next file inside the TAR while r| doesn't do any seeks.
            # r| might be slower but for compressed files we have to go over all the data once anyways.
            # Note that with ignore_zeros = True, no invalid header issues or similar will be raised even for
            # non TAR files!?
            # fmt:off
            return tarfile.open(
                fileobj      = fileObject,
                mode         = 'r|' if self._parent.compression else 'r:',
                ignore_zeros = self._parent.ignoreZeros,
                encoding     = self._parent.encoding,
            )
            # fmt:on
        except tarfile.ReadError:
            pass

        return []

    def _process_serial(self, fileObject: IO[bytes], pathPrefix: str, streamOffset: int) -> Iterable[Tuple]:
        """
        Opens the given fileObject using the tarfile module, iterates over all files converting their metadata to
        FileInfo tuples and inserting those into the database in a chunked manner using the given _set_file_infos.
        """

        if self._recursionDepth > self._parent.maxRecursionDepth:
            return []

        loadedTarFile: Any = self._open_tar(fileObject)

        # Iterate over files inside TAR and add them to the database
        fileInfos: List[Tuple] = []
        xattrs: List[Tuple] = []
        filesToMountRecursively: List[Tuple] = []

        # thread_time is twice as fast, which can shave off 10% of time in some tests but it is not as "correct"
        # because it does not count the sleep time of the thread, e.g., caused by waiting for I/O or even waiting
        # for work done inside multiprocessing.pool.Pool! This can lead to more than factor 10 distortions and
        # therefore is not suitable. If time.time is indeed an issue, then it should be better to use _processParallel.
        self._lastUpdateTime = time.time()

        try:
            for tarInfo in loadedTarFile:
                # Clear this in order to limit memory usage by tarfile.
                loadedTarFile.members = []
                # Note that the .members trick already assumes that we are only iterating over the tar file once.
                # The next iteration would fail because loadedTarFile._loaded would be True and .members would be
                # yielded even though they are now empty. Therefore it is fine to do some monkey-patching here.
                if 'size' in tarInfo.pax_headers:
                    # Patch https://github.com/python/cpython/issues/136601
                    for key in ('GNU.sparse.size', 'GNU.sparse.realsize'):
                        if key in tarInfo.pax_headers:
                            with contextlib.suppress(ValueError):
                                tarInfo.size = int(tarInfo.pax_headers[key])

                    # Patch https://github.com/python/cpython/issues/136602
                    if (
                        tarInfo.pax_headers.get('GNU.sparse.major') == '1'
                        and tarInfo.pax_headers.get('GNU.sparse.minor') == '0'
                        and isinstance(getattr(tarInfo, '_link_target', None), int)
                    ):
                        # pylint: disable=protected-access
                        loadedTarFile.offset = (
                            tarInfo._link_target + ceil_div(int(tarInfo.pax_headers['size']), BLOCK_SIZE) * BLOCK_SIZE
                        )

                # ProgressBar does a similar check like this inside 'update' but doing this outside avoids huge
                # call stacks and also avoids calling tell() on the file object in each loop iteration.
                # I could observe 10% shorter runtimes because of this with the test file:
                #     tar-with-1000-folders-with-1000-files-0B-files.tar
                if time.time() - self._lastUpdateTime >= 1:
                    self._lastUpdateTime = time.time()
                    self._update_progress_bar()

                newFileInfos, newXAttrs, mightBeTar, self._parent._isGnuIncremental = (
                    _TarFileMetadataReader._process_tar_info(
                        tarInfo,
                        fileObject=fileObject,
                        pathPrefix=pathPrefix,
                        streamOffset=streamOffset,
                        isGnuIncremental=self._parent._isGnuIncremental,
                        mountRecursively=self._recursionDepth < self._parent.maxRecursionDepth,
                        transform=self._parent.transform,
                        recursionDepth=self._recursionDepth,
                        printDebug=self._parent.printDebug,
                    )
                )

                if mightBeTar:
                    filesToMountRecursively.extend(newFileInfos)

                fileInfos.extend(newFileInfos)
                if len(fileInfos) > 1000:
                    self._set_file_infos(fileInfos)
                    fileInfos.clear()

                xattrs.extend(newXAttrs)
                if len(xattrs) > 1000:
                    self._setxattrs(xattrs)
                    xattrs.clear()

        finally:
            self._set_file_infos(fileInfos)
            self._setxattrs(xattrs)

        return filesToMountRecursively

    def process(self, fileObject: IO[bytes], pathPrefix: str, streamOffset: int) -> Iterable[Tuple]:
        """
        Iterates over all files inside the given fileObject TAR and inserts their metadata into the database using
        the given _set_file_infos.
        A list of files which might be of interest for recursive mounting of uncompressed TARs is returned.
        """

        try:
            return self._process_serial(fileObject, pathPrefix, streamOffset)

        except tarfile.ReadError as e:
            if 'unexpected end of data' in str(e):
                print(
                    "[Warning] The TAR file is incomplete. Ratarmount will work but some files might be cut off. "
                    "If the TAR file size changes, ratarmount will recreate the index during the next mounting."
                )
                if self._parent.printDebug >= 3:
                    traceback.print_exc()

        return []


class SQLiteIndexedTar(SQLiteIndexMountSource):
    """
    This class reads once through the whole TAR archive and stores TAR file offsets
    for all contained files in an index to support fast seeking to a given file.
    """

    DEFAULT_GZIP_SEEK_POINT_SPACING = 16 * 1024 * 1024

    # fmt: off
    def __init__(
        self,
        tarFileName                  : Optional[Union[str, os.PathLike]] = None,
        fileObject                   : Optional[IO[bytes]]               = None,
        *,  # force all parameters after to be keyword-only
        writeIndex                   : bool                              = False,
        clearIndexCache              : bool                              = False,
        indexFilePath                : Optional[str]                     = None,
        indexFolders                 : Optional[List[str]]               = None,
        recursive                    : bool                              = False,
        gzipSeekPointSpacing         : int                               = DEFAULT_GZIP_SEEK_POINT_SPACING,
        encoding                     : str                               = tarfile.ENCODING,
        stripRecursiveTarExtension   : bool                              = False,
        ignoreZeros                  : bool                              = False,
        verifyModificationTime       : bool                              = False,
        parallelization              : int                               = 1,
        parallelizations             : Optional[Dict[str, int]]          = None,
        isGnuIncremental             : Optional[bool]                    = None,
        printDebug                   : int                               = 0,
        transformRecursiveMountPoint : Optional[Tuple[str, str]]         = None,
        transform                    : Optional[Tuple[str, str]]         = None,
        prioritizedBackends          : Optional[List[str]]               = None,
        indexMinimumFileCount        : int                               = 0,
        recursionDepth               : Optional[int]                     = None,
        # pylint: disable=unused-argument
        **kwargs
    ) -> None:
        """
        tarFileName
            Path to the TAR file to be opened. If not specified, a fileObject must be specified.
            If only a fileObject is given, the created index can't be cached (efficiently).
        fileObject
            A io.IOBase derived object. If not specified, tarFileName will be opened.
            If it is an instance of IndexedBzip2File, IndexedGzipFile, or IndexedZstdFile, then the offset
            loading and storing from and to the SQLite database is managed automatically by this class.
        writeIndex
            If true, then the sidecar index file will be written to a suitable location.
            Will be ignored if indexFilePath is ':memory:' or if only fileObject is specified
            but not tarFileName.
        clearIndexCache
            If true, then check all possible index file locations for the given tarFileName/fileObject
            combination and delete them. This also implicitly forces a recreation of the index.
        indexFilePath
            Path to the index file for this TAR archive. This takes precedence over the automatically
            chosen locations. If it is ':memory:', then the SQLite database will be kept in memory
            and not stored to the file system at any point.
        indexFolders
            Specify one or multiple paths for storing .index.sqlite files. Paths will be tested for
            suitability in the given order. An empty path will be interpreted as the location in which
            the TAR resides.
        recursive
            If true, then TAR files inside this archive will be recursively analyzed and added to the SQLite
            index. Currently, this recursion can only break the outermost compression layer. I.e., a .tar.bz2
            file inside a tar.bz2 file can not be mounted recursively.
        gzipSeekPointSpacing
            This controls the frequency of gzip decoder seek points, see indexed_gzip documentation.
            Larger spacings lead to less memory usage but increase the constant seek overhead.
        encoding
            Will be forwarded to tarfile. Specifies how filenames inside the TAR are encoded.
        ignoreZeros
            Will be forwarded to tarfile. Specifies to not only skip zero blocks but also blocks with
            invalid data. Setting this to true can lead to some problems but is required to correctly
            read concatenated tars.
        stripRecursiveTarExtension
            If true and if recursive is also true, then a <file>.tar inside the current
            tar will be mounted at <file>/ instead of <file>.tar/.
        transformRecursiveMountPoint
            If specified, then a <path>.tar inside the current tar will be matched with the
            first argument of the tuple and replaced by the second argument. This new
            modified path is used as recursive mount point. See also Python's re.sub.
        verifyModificationTime
            If true, then the index will be recreated automatically if the TAR archive has a more
            recent modification time than the index file.
        parallelization
            The amount of parallelization to be used for necessary compression backends if it offers
            parallelization at all and if it makes sense.
        parallelizations
            Fine-granular parallelization for each compression backend. Set the empty string
            key to the default parallelization for backends not explicitly in the dictionary.
            This has higher precedence than 'parallelization'.
        isGnuIncremental
            If None, then it will be determined automatically. Behavior can be overwritten by setting
            it to a bool value. If true, then prefixes will be stripped from certain paths encountered
            with GNU incremental backups.
        kwargs
            Unused. Only for compatibility with generic MountSource interface.
        """

        self.encoding                     = encoding
        self.stripRecursiveTarExtension   = stripRecursiveTarExtension
        self.transformRecursiveMountPoint = transformRecursiveMountPoint
        self.transformPattern             = transform
        self.ignoreZeros                  = ignoreZeros
        self.verifyModificationTime       = verifyModificationTime
        self.gzipSeekPointSpacing         = gzipSeekPointSpacing
        self.printDebug                   = printDebug
        self.isFileObject                 = fileObject is not None
        self._isGnuIncremental            = isGnuIncremental
        self.hasBeenAppendedTo            = False
        self._recursionDepth              = -1
        # fmt: on
        self.prioritizedBackends: List[str] = [] if prioritizedBackends is None else prioritizedBackends
        self.maxRecursionDepth = determine_recursion_depth(recursive=recursive, recursionDepth=recursionDepth)
        self.parallelizations = copy.deepcopy(parallelizations) if parallelizations else {}
        if '' not in self.parallelizations:
            self.parallelizations[''] = parallelization

        self.transform = (
            (lambda x: re.sub(self.transformPattern[0], self.transformPattern[1], x))
            if isinstance(self.transformPattern, (tuple, list)) and len(self.transformPattern) == 2
            else (lambda x: x)
        )

        # Determine an archive file name to show for debug output and as file name inside the mount point for
        # simple non-TAR gzip/bzip2 stream-compressed files.
        self.tarFileName: str
        if fileObject:
            self.tarFileName = str(tarFileName or '<file object>')
        else:
            if tarFileName:
                # Keep the EXACT file path, do not convert to an absolute path, or else we might trigger
                # recursive FUSE calls, which hangs everything!
                self.tarFileName = str(tarFileName)
            else:
                raise RatarmountError("At least one of tarFileName and fileObject arguments should be set!")
        self._fileNameIsURL = re.match('[A-Za-z0-9]*://', self.tarFileName) is not None

        # If no fileObject given, then self.tarFileName is the path to the archive to open.
        self._fileObjectsToCloseOnDel: List[IO[bytes]] = []
        if not fileObject:
            fileObject = open(self.tarFileName, 'rb')
            self._fileObjectsToCloseOnDel.append(fileObject)
        fileObject.seek(0, io.SEEK_END)
        fileSize = fileObject.tell()
        fileObject.seek(0)  # Even if not interested in the file size, seeking to the start might be useful.
        self._archiveFileSize = fileSize

        # rawFileObject : Only set when opening a compressed file and only kept to keep the
        #                 compressed file handle from being closed by the garbage collector.
        # tarFileObject : File object to the uncompressed (or decompressed) TAR file to read actual data out of.
        # compression   : Stores what kind of compression the originally specified TAR file uses.
        # isTar         : Can be false for the degenerated case of only a bz2 or gz file not containing a TAR
        self.compression: Optional[FileFormatID] = None
        self.tarFileObject, self.rawFileObject, self.compression = open_compressed_file(
            fileObject,
            gzipSeekPointSpacing=gzipSeekPointSpacing,
            parallelizations=self.parallelizations,
            enabledBackends=[
                backend for backend, info in COMPRESSION_BACKENDS.items() if info.delegatedArchiveBackend == 'tarfile'
            ],
            prioritizedBackends=self.prioritizedBackends,
            printDebug=self.printDebug,
        )
        self.isTar = might_be_format(self.tarFileObject, FileFormatID.TAR)
        if not self.isTar:
            if printDebug >= 3:
                print(f"[Info] File object {self.tarFileObject} from {self.tarFileName} is not a TAR.")
            if not self.rawFileObject:
                raise RatarmountError(f"File object ({fileObject!s}) could not be opened as a TAR file!")

        if self.compression:
            self._recursionDepth += 1
            # Change the default from 0 to 1 to undo the compression layer and analyze the TAR if nothing is specified.
            if not recursive and recursionDepth is None:
                self.maxRecursionDepth = 1

        # Can only be set correctly after the compression has been detected because it determines the default
        # recursion depth. This is legacy, i.e., only to add the correct value to 'metadata' to be backward compatible.
        self.mountRecursively = self.maxRecursionDepth > 0

        # Try to get block size from the real opened file.
        self.blockSize = 512
        try:
            if self.rawFileObject:
                self.blockSize = os.fstat(self.rawFileObject.fileno()).st_blksize
            elif self.tarFileObject:
                self.blockSize = os.fstat(self.tarFileObject.fileno()).st_blksize
        except Exception:
            pass

        if self.compression == FileFormatID.GZIP:
            self.blockSize = max(self.blockSize, gzipSeekPointSpacing)
        elif self.compression == FileFormatID.BZIP2:
            # There is no API yet to query the exact block size, but most bzip2 files have 900K blocks.
            # The bzip2 block size is in reference to the BWT buffer, so the decompressed block data will be
            # larger to some extend. Therefore, 1 MiB blocks should be an alright guess for all bzip2 files.
            self.blockSize = 1024 * 1024
        # TODO derive some meaningful block sizes for zstd and xz

        self.fileObjectLock = threading.Lock()

        if (
            self.compression == FileFormatID.XZ
            and len(getattr(self.tarFileObject, 'block_boundaries', (0, 0))) <= 1
            and self._archiveFileSize > 1024 * 1024
        ):
            print(f"[Warning] The specified file '{self.tarFileName}'")
            print("[Warning] is compressed using xz but only contains one xz block. This makes it ")
            print("[Warning] impossible to use true seeking! Please (re)compress your TAR using pixz")
            print("[Warning] (see https://github.com/vasi/pixz) in order for ratarmount to do be able ")
            print("[Warning] to do fast seeking to requested files.")
            print("[Warning] As it is, each file access will decompress the whole TAR from the beginning!")
            print()

        if indexFolders is None:
            indexFolders = ['', os.path.join("~", ".ratarmount")]
            xdgCacheHome = get_xdg_cache_home()
            if xdgCacheHome and os.path.isdir(os.path.expanduser(xdgCacheHome)):
                indexFolders.insert(1, os.path.join(xdgCacheHome, 'ratarmount'))
        elif isinstance(indexFolders, str):
            indexFolders = [indexFolders]

        archiveFilePath = self.tarFileName if not self.isFileObject or self._fileNameIsURL else None

        super().__init__(
            SQLiteIndex(
                indexFilePath,
                indexFolders=indexFolders,
                archiveFilePath=archiveFilePath,
                encoding=self.encoding,
                printDebug=self.printDebug,
                indexMinimumFileCount=indexMinimumFileCount,
                backendName='SQLiteIndexedTar',
                ignoreCurrentFolder=self.isFileObject and self._fileNameIsURL,
            ),
            clearIndexCache=clearIndexCache,
            checkMetadata=self._check_metadata,
        )

        if self.index.index_is_loaded():
            if not self.hasBeenAppendedTo:  # indirectly set by a successful call to _try_load_index
                self._load_or_store_compression_offsets()  # load
                self.index.reload_index_read_only()
                return

            # TODO Handling appended files to compressed archives would have to account for dropping the offsets,
            #      seeking to the first appended file while not processing any metadata and still showing a progress
            #      bar as well as saving the block offsets out after reading and possibly other things.
            if self.compression:
                # When loading compression offsets, the backends assume they are complete, so we have to clear them.
                self.index.clear_compression_offsets()

            pastEndOffset = self._get_past_end_offset(self.index.get_connection())
            if not self.compression and pastEndOffset and self._check_index_validity():
                archiveSize = self.tarFileObject.seek(0, io.SEEK_END)

                newShare = (archiveSize - pastEndOffset) / archiveSize
                print(f"Detected TAR being appended to. Will only analyze the newly added {newShare:.2f} % of data.")

                appendedPartAsFile = StenciledFile(
                    fileStencils=[(self.tarFileObject, pastEndOffset, archiveSize - pastEndOffset)]
                )
                self._create_index(appendedPartAsFile, streamOffset=pastEndOffset)

                self._load_or_store_compression_offsets()  # store

                self.index.drop_metadata()
                self._store_metadata()
                self.index.reload_index_read_only()
                return

            self.index.close()
            print("[Warning] The loaded index does not match the archive. Will recreate it.")

        # TODO This does and did not work correctly for recursive TARs because the outermost layer will change
        #      None to a hard value and from then on it would have been fixed to that value even when called
        #      inside createIndex.
        # Required for _create_index
        if self._isGnuIncremental is None:
            self._isGnuIncremental = self._detect_gnu_incremental(self.tarFileObject)

        # Open new database when we didn't find an existing one.
        if not self.index.index_is_loaded():
            # Simply open in memory without an error even if writeIndex is True but when no indication
            # for an index file location has been given.
            if writeIndex and (indexFilePath or self._get_archive_path() or not self.isFileObject):
                self.index.open_writable()
            else:
                self.index.open_in_memory()

        self._create_index(self.tarFileObject)
        self._load_or_store_compression_offsets()  # store
        if self.index.index_is_loaded():
            self._store_metadata()
            self.index.reload_index_read_only()

        if self.printDebug >= 1 and self.index.indexFilePath and os.path.isfile(self.index.indexFilePath):
            # The 0-time is legacy for the automated tests
            # fmt: off
            print("Writing out TAR index to", self.index.indexFilePath, "took 0s",
                  "and is sized", os.stat( self.index.indexFilePath ).st_size, "B")
            # fmt: on

    def __del__(self):
        if hasattr(self, '_fileObjectsToCloseOnDel'):
            for fileObject in self._fileObjectsToCloseOnDel:
                close = getattr(fileObject, 'close', None)
                if close and callable(close):
                    close()
        if hasattr(super(), '__del__'):
            super().__del__()

    def _detect_gnu_incremental(self, fileObject: Any) -> bool:
        """Check for GNU incremental backup TARs."""
        oldPos = fileObject.tell()

        t0 = time.time()
        try:
            # For an uncompressed 500MB TAR, this iteration took ~0.7s for 1M files roughly 30x faster than tarfile.
            # But for compressed TARs or for HDDs as opposed to SSDs, this might be much slower.
            nMaxToTry = 1000 if self.isFileObject or self.compression else 10000
            for _, typeFlag in _TarFileMetadataReader.find_tar_file_offsets(fileObject, self.ignoreZeros):
                # It seems to be possible to create mixtures of incremental archives and normal contents,
                # therefore do not check that all files must have the mtime prefix.
                if typeFlag == b'D':
                    if self.printDebug >= 1:
                        print("[Info] Detected GNU incremental TAR.")
                    return True

                nMaxToTry -= 1
                if nMaxToTry <= 0 or time.time() - t0 > 3:
                    break

        except Exception as exception:
            if self.printDebug >= 3:
                print("[Info] TAR was not recognized as GNU incremental TAR because of exception", exception)
            if self.printDebug >= 3:
                traceback.print_exc()
        finally:
            fileObject.seek(oldPos)

        return False

    def close(self):
        if self.tarFileObject:
            self.tarFileObject.close()

        if not self.isFileObject and self.rawFileObject:
            self.rawFileObject.close()

    @overrides(MountSource)
    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()
        super().__exit__(exception_type, exception_value, exception_traceback)

    def _get_archive_path(self) -> Optional[str]:
        return None if self.tarFileName == '<file object>' else self.tarFileName

    def _store_metadata(self) -> None:
        argumentsToSave = [
            'mountRecursively',
            'gzipSeekPointSpacing',
            'encoding',
            'stripRecursiveTarExtension',
            'transformRecursiveMountPoint',
            'transformPattern',
            'ignoreZeros',
        ]

        argumentsMetadata = json.dumps({argument: getattr(self, argument) for argument in argumentsToSave})
        # The second argument must be a path to a file to call os.stat with, not simply a file name.
        self.index.store_metadata(argumentsMetadata, None if self.isFileObject else self.tarFileName)
        self.index.store_metadata_key_value('isGnuIncremental', '1' if self._isGnuIncremental else '0')

    def _update_progress_bar(self, progressBar, fileobj: Any) -> None:
        if not progressBar:
            return

        try:
            if (
                hasattr(fileobj, 'tell_compressed')
                and 'rapidgzip' in sys.modules
                and isinstance(fileobj, (rapidgzip.IndexedBzip2File, rapidgzip.RapidgzipFile))
            ):
                # Note that because bz2 works on a bitstream the tell_compressed returns the offset in bits
                progressBar.update(fileobj.tell_compressed() // 8)
            elif hasattr(fileobj, 'tell_compressed'):
                progressBar.update(fileobj.tell_compressed())
            elif hasattr(fileobj, 'fileobj') and callable(fileobj.fileobj):
                progressBar.update(fileobj.fileobj().tell())
            elif isinstance(fileobj, ParallelXZReader):
                blockNumber = fileobj.find_block(fileobj.tell())
                if blockNumber and blockNumber < len(fileobj.approximateCompressedBlockBoundaries):
                    progressBar.update(fileobj.approximateCompressedBlockBoundaries[blockNumber])
            elif self.rawFileObject and hasattr(self.rawFileObject, 'tell'):
                progressBar.update(self.rawFileObject.tell())
            else:
                progressBar.update(fileobj.tell())
        except Exception as exception:
            if self.printDebug >= 1:
                print("An exception occurred when trying to update the progress bar:", exception)
            if self.printDebug >= 3:
                traceback.print_exc()

    def _create_index(self, fileObject: IO[bytes], streamOffset: int = 0) -> None:
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.tarFileName} ...")
        t0 = timer()

        self.index.ensure_intermediary_tables()

        progressBar = ProgressBar(self._archiveFileSize)
        self._create_index_recursively(
            fileObject,
            progressBar=progressBar,
            pathPrefix="",
            streamOffset=streamOffset,
            recursionDepth=self._recursionDepth + 1,
        )

        # Resort by (path,name). This one-time resort is faster than resorting on each INSERT (cache spill)
        if self.printDebug >= 2:
            print("Resorting files by path ...")

        self.index.finalize()

        t1 = timer()
        if self.printDebug >= 1:
            print(f"Creating offset dictionary for {self.tarFileName} took {t1 - t0:.2f}s")

    def _create_index_recursively(
        self, fileObject: IO[bytes], progressBar: ProgressBar, pathPrefix: str, streamOffset: int, recursionDepth: int
    ) -> None:
        metadataReader = _TarFileMetadataReader(
            self,
            self.index.set_file_infos,
            self.index.setxattrs,
            lambda: self._update_progress_bar(progressBar, fileObject),
            recursionDepth=recursionDepth,
        )
        filesToMountRecursively = metadataReader.process(fileObject, pathPrefix, streamOffset)

        # 4. Open contained TARs for recursive mounting
        oldPos = fileObject.tell()
        oldPrintName = self.tarFileName
        for fileInfo in filesToMountRecursively:
            modifiedFolder = fileInfo[0]
            modifiedName = fileInfo[1]

            # Strip file extension for mount point if so configured
            tarExtension = '.tar'
            if (
                self.stripRecursiveTarExtension
                and len(tarExtension) > 0
                and modifiedName.lower().endswith(tarExtension.lower())
            ):
                modifiedName = modifiedName[: -len(tarExtension)]

            # Apply regex transformation to get mount point
            pattern = self.transformRecursiveMountPoint
            modifiedPath = '/' + (modifiedFolder + '/' + modifiedName).lstrip('/')
            if isinstance(pattern, (tuple, list)) and len(pattern) == 2:
                modifiedPath = '/' + re.sub(pattern[0], pattern[1], modifiedPath).lstrip('/')
                modifiedFolder, modifiedName = modifiedPath.rsplit('/', 1)

            # Temporarily change tarFileName for the info output of the recursive call
            self.tarFileName = os.path.join(fileInfo[0], fileInfo[1])

            # StenciledFile's tell returns the offset inside the file chunk instead of the global one,
            # so we have to always communicate the offset of this chunk to the recursive call no matter
            # whether tarfile has streaming access or seeking access!
            globalOffset = fileInfo[3]
            size = fileInfo[4]
            # fileObject already effectively applies streamOffset, so we can't use the globalOffset here!
            # For all supported cases, it should be fine to directly use self.tarFileObject instead of fileObject.
            # This would also save some indirections to speed up accesses.
            tarFileObject = StenciledFile([(self.tarFileObject, globalOffset, size)])

            isTar = False
            try:
                # Do not use os.path.join here because the leading / might be missing.
                # This should instead be seen as the reverse operation of the rsplit further above.
                self._create_index_recursively(
                    tarFileObject,
                    progressBar,
                    modifiedPath,
                    streamOffset=globalOffset,
                    recursionDepth=recursionDepth + 1,
                )
                isTar = True
            except tarfile.ReadError:
                pass
            finally:
                del tarFileObject

            if isTar:
                modifiedFileInfo = list(fileInfo)

                # if the TAR file contents could be read, we need to adjust the actual
                # TAR file's metadata to be a directory instead of a file.
                # Avoid overwriting that data, instead add new one such that it can be versioned
                # or ignored on depending on the recursion depth!
                mode = modifiedFileInfo[6]
                mode = (
                    (mode & 0o777)
                    | stat.S_IFDIR
                    | (stat.S_IXUSR if mode & stat.S_IRUSR != 0 else 0)
                    | (stat.S_IXGRP if mode & stat.S_IRGRP != 0 else 0)
                    | (stat.S_IXOTH if mode & stat.S_IROTH != 0 else 0)
                )

                if modifiedFolder != modifiedFileInfo[0] or modifiedName != modifiedFileInfo[1]:
                    modifiedFileInfo[0] = modifiedFolder
                    modifiedFileInfo[1] = modifiedName
                else:
                    # Increment offset and offsetheader such that the new folder is seen as a more recent version
                    # of the already existing file path for the archive if it has the same path. Else, it would
                    # be undetermined which version is to be counted as more recent when using ORDER BY offsetheader.
                    # Note that offset and offsetheader contain a lot of redundant bits anyway because they are known
                    # to be 0 modulo 512, so the original offsets can be reconstructed even after adding 1.
                    modifiedFileInfo[2] = modifiedFileInfo[2] + 1
                    modifiedFileInfo[3] = modifiedFileInfo[3] + 1
                modifiedFileInfo[6] = mode
                modifiedFileInfo[11] = isTar
                modifiedFileInfo[13] = True  # is generated, i.e., does not have xattr
                modifiedFileInfo[14] += 1  # recursion depth

                self.index.set_file_info(tuple(modifiedFileInfo))

                # Update isTar to True for the tar
                modifiedFileInfo = list(fileInfo)
                modifiedFileInfo[11] = isTar

                self.index.set_file_info(tuple(modifiedFileInfo))

        fileObject.seek(oldPos)
        self.tarFileName = oldPrintName

        # If no file is in the TAR, then it most likely indicates a possibly compressed non TAR file.
        # In that case add that itself to the file index. This will be ignored when called recursively
        # because the table will at least contain the recursive file to mount itself, i.e., file_count > 0
        if self.index.file_count() == 0:
            if self.printDebug >= 3:
                print(f"Did not find any file in the given TAR: {self.tarFileName}. Assuming a compressed file.")

            # For some reason, this happens for single-file.iso.
            # Tarfile does not raise an error but also does not find any files.
            if not self.compression:
                raise CompressionError("Tarfile returned nothing, not even an error, and the file is not compressed!")

            tarInfo: Optional[Any] = None
            try:
                tarInfo = os.fstat(fileObject.fileno())
            except io.UnsupportedOperation:
                # If fileObject doesn't have a fileno, we set tarInfo to None
                # and set the relevant statistics (such as st_mtime) to sensible defaults.
                tarInfo = None

            fname = os.path.basename(self.tarFileName)
            for suffix in ['.gz', '.bz2', '.bzip2', '.gzip', '.xz', '.zst', '.zstd', '.zz', '.zlib']:
                if fname.lower().endswith(suffix) and len(fname) > len(suffix):
                    fname = fname[: -len(suffix)]
                    break

            # Try to get original file name from gzip
            mtime = 0
            if self.rawFileObject:
                oldPos = self.rawFileObject.tell()
                self.rawFileObject.seek(0)
                try:
                    info = get_gzip_info(self.rawFileObject)
                    if info:
                        fname, mtime = info
                except Exception:
                    if self.printDebug >= 2:
                        print("[Info] Could not determine an original gzip file name probably because it is not a gzip")
                    if self.printDebug >= 3:
                        traceback.print_exc()
                finally:
                    # TODO Why does tell return negative numbers!? Problem with indexed_gzip?
                    self.rawFileObject.seek(max(0, oldPos))

            # If the file object is actually an IndexedBzip2File or such, we can't directly use the file size
            # from os.stat and instead have to gather it from seek. Unfortunately, indexed_gzip does not support
            # io.SEEK_END even though it could as it has the index ...
            while fileObject.read(1024 * 1024):
                self._update_progress_bar(progressBar, fileObject)
            fileSize = fileObject.tell()

            mode = 0o777 | stat.S_IFREG  # default mode

            # fmt: off
            fileInfo = (
                ""                                    ,  # 0  : path
                fname                                 ,  # 1  : file name
                None                                  ,  # 2  : header offset
                0                                     ,  # 3  : data offset
                fileSize                              ,  # 4  : file size
                tarInfo.st_mtime if tarInfo else mtime,  # 5  : modification time
                tarInfo.st_mode if tarInfo else mode  ,  # 6  : file mode / permissions
                None                                  ,  # 7  : TAR file type. Currently unused. Overlaps with mode
                None                                  ,  # 8  : linkname
                tarInfo.st_uid if tarInfo else 0      ,  # 9  : user ID
                tarInfo.st_gid if tarInfo else 0      ,  # 10 : group ID
                False                                 ,  # 11 : isTar
                False,  # 12 isSparse, don't care if it is actually sparse or not because it is not in TAR
                False                                 ,  # 13 : is generated (parent folder)
                recursionDepth - 1                    ,  # 14 : recursion depth
            )
            # fmt: on
            self.index.set_file_info(fileInfo)

    def _open_stencil(self, offset: int, size: int, buffering: int) -> IO[bytes]:
        if buffering == 0:
            return cast(IO[bytes], RawStenciledFile([(self.tarFileObject, offset, size)], self.fileObjectLock))
        return cast(
            IO[bytes],
            StenciledFile(
                [(self.tarFileObject, offset, size)],
                self.fileObjectLock,
                bufferSize=self.blockSize if buffering == -1 else buffering,
            ),
        )

    @overrides(MountSource)
    def open(self, fileInfo: FileInfo, buffering=-1) -> IO[bytes]:
        assert fileInfo.userdata
        tarFileInfo = fileInfo.userdata[-1]
        assert isinstance(tarFileInfo, SQLiteIndexedTarUserData)

        # This is not strictly necessary but it saves two file object layers and therefore might be more performant.
        # Furthermore, non-sparse files should be the much more likely case anyway.
        if not tarFileInfo.issparse:
            return self._open_stencil(tarFileInfo.offset, fileInfo.size, buffering)

        # The TAR file format is very simple. It's just a concatenation of TAR blocks. There is not even a
        # global header, only the TAR block headers. That's why we can simply cut out the TAR block for
        # the sparse file using StenciledFile and then use tarfile on it to expand the sparse file correctly.
        tarBlockSize = tarFileInfo.offset - tarFileInfo.offsetheader + fileInfo.size

        tarSubFile = self._open_stencil(tarFileInfo.offsetheader, tarBlockSize, buffering)
        # TODO It might be better to somehow call close on tarFile but the question is where and how.
        #      It would have to be appended to the __exit__ method of fileObject like if being decorated.
        #      For now this seems to work either because fileObject does not require tarFile to exist
        #      or because tarFile is simply not closed correctly here, I'm not sure.
        #      Sparse files are kinda edge-cases anyway, so it isn't high priority as long as the tests work.
        tarFile = tarfile.open(fileobj=cast(IO[bytes], tarSubFile), mode='r:', encoding=self.encoding)
        tarInfo = next(iter(tarFile))

        # Patch https://github.com/python/cpython/issues/136601
        if 'size' in tarInfo.pax_headers:
            for key in ('GNU.sparse.size', 'GNU.sparse.realsize'):
                if key in tarInfo.pax_headers:
                    with contextlib.suppress(ValueError):
                        tarInfo.size = int(tarInfo.pax_headers[key])

        fileObject = tarFile.extractfile(tarInfo)
        if not fileObject:
            raise CompressionError("tarfile.extractfile returned nothing!")

        return fileObject

    @overrides(MountSource)
    def read(self, fileInfo: FileInfo, size: int, offset: int) -> bytes:
        assert fileInfo.userdata
        tarFileInfo = fileInfo.userdata[-1]
        assert isinstance(tarFileInfo, SQLiteIndexedTarUserData)

        if tarFileInfo.issparse:
            with self.open(fileInfo) as file:
                file.seek(offset, os.SEEK_SET)
                return file.read(size)

        # For non-sparse files, we can simply seek to the offset and read from it.
        self.tarFileObject.seek(tarFileInfo.offset + offset, os.SEEK_SET)
        return self.tarFileObject.read(size)

    @overrides(MountSource)
    def statfs(self) -> Dict[str, Any]:
        return {
            'f_bsize': self.blockSize,
            'f_frsize': self.blockSize,
            'f_bfree': 0,
            'f_bavail': 0,
            'f_ffree': 0,
            'f_favail': 0,
        }

    @staticmethod
    def _get_past_end_offset(sqlConnection: sqlite3.Connection) -> Optional[int]:
        """
        Returns None if it cannot determine where the archive should end. Currently, because of implementation
        limitations, this may happen if the last entry in the archive is a sparse file.
        """
        # TODO Make it work with sparse files by analyzing those sparse blocks manually or maybe get tarfile to do it

        # Note that we cannot use the recorded archive file size to determine from which we need to resume
        # reading because it is not specified how many zero-byte blocks there may be at the end:
        # > At the end of the archive file there shall be two 512-byte blocks filled with binary zeros,
        # > interpreted as an end-of-archive indicator.
        # For example, GNU tar rounds up to 10 KiB for very small archives but will (have to) append further
        # files right after the the last non-zero block, which might be at offset 512 for empty files.
        # > The user can specify a blocking factor, which is the number of blocks per record.
        # > The default is 20, producing 10 KiB records.
        result = sqlConnection.execute(
            "SELECT offset + size, issparse FROM files ORDER BY offset DESC LIMIT 1"
        ).fetchone()
        if not result:
            raise InvalidIndexError("The index contains no files!")
        pastEndOffset, isSparse = result

        if isSparse:
            return None

        # Round up to next TAR block
        if pastEndOffset % 512 != 0:
            pastEndOffset += 512 - (pastEndOffset % 512)

        return pastEndOffset

    def _try_to_mark_as_appended(self, storedStats: Dict[str, Any], archiveStats: os.stat_result):
        """
        Raises an exception if it makes no sense to only try to go over the new appended data alone
        else sets self.hasBeenAppendedTo to True.
        There is one very specific usecase for which recreating the complete index would be a waste:
        When an uncompressed archive got appended a rather small amount of files.
        """

        # Sizes should be determined and larger or equal
        if (
            not hasattr(archiveStats, 'st_size')
            or 'st_size' not in storedStats
            or archiveStats.st_size < storedStats['st_size']
        ):
            raise InvalidIndexError(
                "Will not treat an archive that shrank or has indeterminable size as having been appended to!"
            )

        # Times should be determined and larger or equal
        if (
            not hasattr(archiveStats, "st_mtime")
            or 'st_mtime' not in storedStats
            or archiveStats.st_mtime < storedStats['st_mtime']
        ):
            # Always throw even for if self.verifyModificationTime is False because in this method,
            # the archive should already have been determines as different.
            raise InvalidIndexError(
                f"The modification date for the TAR file {storedStats['st_mtime']} "
                f"is older than the one stored in the SQLite index ({archiveStats.st_mtime!s})",
            )

        # Checking is expensive and would basically do the same work as creating the database anyway.
        # Therefore, only bother with the added complexity and uncertainty of the randomized index check
        # if the additional part to analyze makes up less than 66% of the total archive.
        #
        # Ignore small archives that don't require much time to process anyway.
        # The threshold is motivated by the benchmarks for "First Mounting".
        # For uncompressed archives, the limiting factor is the number of files.
        # An uncompressed TAR with 1000 64KiB files would take roughly a second.
        if archiveStats.st_size < 64 * 1024 * 1024:
            raise InvalidIndexError("The archive did change but is too small to determine as having been appended to.")

        if self.index.file_count() < SQLiteIndex.NUMBER_OF_METADATA_TO_VERIFY:
            raise InvalidIndexError(
                "The archive did change but has too few files to determine as having been appended to."
            )

        # If the archive more than tripled, then the already existing part isn't all that much in
        # comparison to the work that would have to be done anyway. And because the validity check
        # would have to only be an approximation, simply allow the up to 33% overhead to recreate
        # everything from scratch, just to be sure.
        if archiveStats.st_size > 3 * storedStats['st_size']:
            raise InvalidIndexError(
                f"TAR file for this SQLite index has more than tripled in size from "
                f"{storedStats['st_size']} to {archiveStats.st_size}"
            )

        # Note that the xz compressed version of 100k zero-byte files is only ~200KB!
        # But this should be an edge-case and with a compression ratio of ~2, even compressed archives
        # of this size should not take more than 10s, so pretty negligible in my opinion.
        #
        # For compressed archives, detecting appended archives does not help much because the bottleneck is
        # the decompression not the indexing of files. And because rapidgzip and indexed_gzip probably
        # assume that the index is complete once import_index has been called, we have to recreate the full
        # block offsets anyway.
        if self.compression:
            raise InvalidIndexError(
                f"Compressed TAR file for this SQLite index has changed size from "
                f"{storedStats['st_size']} to {archiveStats.st_size}. It cannot be treated as appended."
            )

        if self.index.get_index_version() != SQLiteIndex.__version__:
            raise InvalidIndexError("Cannot append to index of different versions!")

        if self.printDebug >= 2:
            print("[Info] Archive has probably been appended to because it is larger and more recent.")
        self.hasBeenAppendedTo = True

    def _check_metadata(self, metadata: Dict[str, Any]) -> None:
        # self._isGnuIncremental may be initialized during metadata check because it is required for some checks.
        # But, if the subsequent checks fail, then we want to restore the initial value.
        isGnuIncremental = self._isGnuIncremental
        try:
            self._check_metadata2(metadata)
        except Exception as e:
            self._isGnuIncremental = isGnuIncremental
            raise e

    def _check_metadata2(self, metadata: Dict[str, Any]) -> None:
        """
        Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible.
        Returns normally and sets self.index.hasBeenAppendedTo to True if the size of the archive increased
        but still fits.
        """

        if 'tarstats' in metadata:
            storedStats = json.loads(metadata['tarstats'])
            tarStats = os.stat(self.tarFileName)

            if hasattr(tarStats, 'st_size') and 'st_size' in storedStats:
                if tarStats.st_size < storedStats['st_size']:
                    raise InvalidIndexError(
                        f"TAR file for this SQLite index has shrunk in size from "
                        f"{storedStats['st_size']} to {tarStats.st_size}"
                    )

                if tarStats.st_size > storedStats['st_size']:
                    self._try_to_mark_as_appended(storedStats, tarStats)

            # For compressed files, the archive size check should be sufficient because even if the uncompressed
            # size does not change, the compressed size will most likely change.
            # And also it would be expensive to do because the block offsets are not yet loaded yet!
            db = self.index.sqlConnection
            pastEndOffset = self._get_past_end_offset(db) if db else None
            if not self.compression and pastEndOffset:
                # https://pubs.opengroup.org/onlinepubs/9699919799/utilities/pax.html#tag_20_92_13_01
                # > At the end of the archive file there shall be two 512-byte blocks filled with binary zeros,
                # > interpreted as an end-of-archive indicator.
                fileStencil = (self.tarFileObject, pastEndOffset, 1024)
                oldOffset = self.tarFileObject.tell()
                try:
                    with RawStenciledFile(fileStencils=[fileStencil]) as file:
                        if file.read(1025) != b"\0" * 1024:
                            if self.printDebug >= 2:
                                print(
                                    "[Info] Probably has been appended to because no EOF zero-byte blocks could "
                                    f"be found at offset: {pastEndOffset}"
                                )
                            self._try_to_mark_as_appended(storedStats, tarStats)
                finally:
                    self.tarFileObject.seek(oldOffset)

            # Only happens very rarely, e.g., for more recent files with the same size.
            if (
                not self.hasBeenAppendedTo
                and self.verifyModificationTime
                and hasattr(tarStats, "st_mtime")
                and 'st_mtime' in storedStats
                and tarStats.st_mtime != storedStats['st_mtime']
            ):
                raise InvalidIndexError(
                    f"The modification date for the TAR file {storedStats['st_mtime']} "
                    f"to this SQLite index has changed ({tarStats.st_mtime!s})",
                )

        # Check arguments used to create the found index.
        # These are only warnings and not forcing a rebuild by default.
        # TODO: Add --force options?
        if 'arguments' in metadata:
            indexArgs = json.loads(metadata['arguments'])
            argumentsToCheck = [
                'mountRecursively',
                'encoding',
                'stripRecursiveTarExtension',
                'transformRecursiveMountPoint',
                'transformPattern',
                'ignoreZeros',
            ]

            if self.compression == FileFormatID.GZIP:
                argumentsToCheck.append('gzipSeekPointSpacing')

            SQLiteIndex.check_metadata_arguments(indexArgs, self, argumentsToCheck)

        # Restore the self._isGnuIncremental flag before doing any row validation because else there could be
        # false positive warnings regarding GNU incremental detection.
        if 'isGnuIncremental' in metadata:
            value = metadata['isGnuIncremental'].lower()
            self._isGnuIncremental = value in ('true', '1')
        elif self.index.sqlConnection:
            # This can be expensive, but it should still be less expensive than rereading the first 1000 file headers
            # and checking the type through that way. There will be a breakeven point though for very large archives.
            # Then, it would be better to update the index to contain the 'isGnuIncremental' metadata key.
            self._isGnuIncremental = bool(
                self.index.sqlConnection.execute(
                    """SELECT 1 FROM "files" WHERE hex(type) = hex("D") LIMIT 1"""
                ).fetchone()
            )

        if 'backendName' not in metadata:
            # Checking the first two should already be enough to detect an index created with a different backend.
            # Do not verify folders because parent folders and root get automatically added!
            result = self.index.get_connection().execute(
                f"""SELECT * {SQLiteIndex.FROM_REGULAR_FILES} ORDER BY offset ASC LIMIT 2;"""
            )
            if not self._check_rows_validity(result):
                raise InvalidIndexError("The first two files of the index do not match.")

    def _check_index_validity(self) -> bool:
        # Check some of the first and last files in the archive and some random selection in between.
        selectFiles = "SELECT * " + SQLiteIndex.FROM_REGULAR_FILES
        result = self.index.get_connection().execute(
            f"""
            SELECT * FROM ( {selectFiles} ORDER BY offset ASC LIMIT 100 )
            UNION
            SELECT * FROM ( {selectFiles} ORDER BY RANDOM() LIMIT {SQLiteIndex.NUMBER_OF_METADATA_TO_VERIFY} )
            UNION
            SELECT * FROM ( {selectFiles} ORDER BY offset DESC LIMIT 100 )
            ORDER BY offset
        """
        )
        return self._check_rows_validity(result)

    def _check_rows_validity(self, rows) -> bool:
        t0 = time.time()

        oldOffset = self.tarFileObject.tell()
        rowCount = 0
        try:
            for row in rows:
                rowCount += 1

                # As for the stencil size, 512 B (one TAR block) would be enough for most cases except for
                # features like GNU LongLink which store additional metadata in further TAR blocks.
                offsetHeader = int(row[2])
                offsetData = int(row[3])
                headerBlockCount = max(1, math.ceil((offsetData - offsetHeader) / 512)) * 512
                with StenciledFile(
                    fileStencils=[(self.tarFileObject, offsetHeader, headerBlockCount)]
                ) as file, tarfile.open(fileobj=file, mode='r|', ignore_zeros=True, encoding=self.encoding) as archive:
                    tarInfo = next(iter(archive))
                    realFileInfos, _, _, _ = _TarFileMetadataReader._process_tar_info(
                        tarInfo,
                        file,  # only used for isGnuIncremental == True
                        "",  # pathPrefix
                        offsetHeader,  # will be added to all offsets to get the real offset
                        isGnuIncremental=self._isGnuIncremental,
                        mountRecursively=False,
                        transform=self.transform,
                        recursionDepth=0,
                        printDebug=self.printDebug,
                    )

                    if not realFileInfos:
                        return False
                    realFileInfo = realFileInfos[0]

                    # Bool columns will have been converted to int 0 or 1 when reading from SQLite.
                    # In order to compare with the read result correctly, we need to convert them to bool, too.
                    storedFileInfo = list(row)
                    for index in [-1, -2]:
                        if storedFileInfo[index] not in [0, 1]:
                            return False
                        storedFileInfo[index] = bool(storedFileInfo[index])

                    # Do not compare the path because it might have the parent TAR prepended to it for
                    # recursive TARs and this is hard to ignore any other way.
                    storedFileInfo[0] = realFileInfo[0]  # path
                    storedFileInfo[11] = realFileInfo[11]  # isTar

                    commonSize = min(len(storedFileInfo), len(realFileInfo))
                    # Do not check newly added columns such as isgenerated and recursiondepth.
                    if commonSize > 13:
                        storedFileInfo[13] = realFileInfo[13]  # is generated
                    if commonSize > 14:
                        storedFileInfo[14] = realFileInfo[14]  # recursion depth
                    if storedFileInfo[:commonSize] != list(realFileInfo)[:commonSize]:
                        if self.printDebug >= 3:
                            print("[Info] Stored file info:")
                            print("[Info]", storedFileInfo)
                            print("[Info] differs from recomputed one:")
                            print("[Info]", realFileInfo)
                        return False

            return True
        except tarfile.TarError:
            # Not even worth warning because this simply might happen if the index is not valid anymore.
            return False
        finally:
            self.tarFileObject.seek(oldOffset)

            if self.printDebug >= 2:
                t1 = time.time()
                print(f"[Info] Verifying metadata for {rowCount} files took {t1 - t0:.3f} s")

        return False

    def _load_or_store_compression_offsets(self):
        if self.compression:
            self.index.synchronize_compression_offsets(self.tarFileObject, self.compression)

    def join_threads(self):
        if hasattr(self.tarFileObject, 'join_threads'):
            self.tarFileObject.join_threads()



"""
This module offers a MountSource interface, which has methods for listing paths
and getting file metadata and contents. File lookup returns a FileInfo object,
which uniquely identifies the file, similar to a filesystem inode, and can be
used to open the file.

There are multiple implementations of the MountSource interface, largely split
into two submodules: "formats" and "compositing".

"formats" are MountSource implementations that work on an input file object or
URL and exposes its file structure and contained files:

 - SQLiteIndexedTar: This is the oldest and most powerful implementation.
                     It supports fast access to files inside (compressed) TARs.
 - RarMountSource: An implementation for RARs using rarfile.
 - ZipMountSource: An implementation for ZIPs using zipfile.
 - FolderMountSource: An implementation taking an existing folder as input.

The "compositing" submodule contains MountSource implementations that offer
higher-level abstractions or functionalities on top of one or more input
MountSource implementation.

 - UnionMountSource: Takes multiple MountSource implementations and merges them.
 - SubvolumesMountSource: Takes multiple MountSource implementations and mounts
                          each in separate subfolders with specified names.
 - FileVersionLayer: Takes a MountSource as input, decodes the requested
                     paths, also accepting "<file>.version/<number>" paths,
                     and calls the methods of the MountSource with the given
                     file version.
 - AutoMountLayer: Takes one MountSource, goes over all its files and mounts
                   archives recursively in a similar manner to UnionMountSource.

The factory function 'open_mount_source' opens one of the format MountSource
implementations according to the file type.
For performance and maintenance reasons, there are almost no module-level
reimports, ergo you should specify the full module hierarchy for imports.

Example:

    from ratarmountcore.mountsource.formats.tar import SQLiteIndexedTar
    from ratarmountcore.mountsource.factory import open_mount_source

    archive = SQLiteIndexedTar("foo.tar")
    # or alternatively:
    archive = open_mount_source("foo.tar", recursive=True)

    archive.list("/")
    info = archive.lookup("/bar")

    print "Contents of /bar:"
    with archive.open(info) as file:
        print(file.read())
"""

from .MountSource import FileInfo, MountSource, create_root_file_info, merge_statfs



# pylint: disable=wrong-import-order
# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import stat
import sys

from helpers import copy_test_file

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.mountsource.formats.py7zr import Py7zrMountSource, py7zr  # noqa: E402


class TestPy7zrMountSource:
    @staticmethod
    def test_password():
        if py7zr is None or not getattr(py7zr, '__version__', '').startswith("1.0"):
            return

        with copy_test_file('encrypted-nested-tar.7z') as path, Py7zrMountSource(
            path, passwords=[b'foo']
        ) as mountSource:
            for folder in ['/', '/foo', '/foo/fighter']:
                fileInfo = mountSource.lookup(folder)
                assert fileInfo
                assert stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(folder) == 1
                assert mountSource.list(folder)

            for filePath in ['/foo/fighter/ufo']:
                fileInfo = mountSource.lookup(filePath)
                assert fileInfo
                assert not stat.S_ISDIR(fileInfo.mode)

                assert mountSource.versions(filePath) == 1
                assert not mountSource.list(filePath)
                with mountSource.open(mountSource.lookup(filePath)) as file:
                    assert file.read() == b'iriya\n'



import contextlib
import json
import os
import re
import shutil
import sqlite3
import stat
import sys
import tarfile
import tempfile
import time
import traceback
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from typing import IO, Any, AnyStr, Callable, Dict, List, Optional, Tuple, Union

try:
    import fsspec
except ImportError:
    fsspec = None  # type: ignore

with contextlib.suppress(ImportError):
    import indexed_gzip

with contextlib.suppress(ImportError):
    import rapidgzip

from .compressions import COMPRESSION_BACKENDS, detect_compression, find_available_backend
from .formats import FILE_FORMATS, FileFormatID
from .mountsource import FileInfo, create_root_file_info
from .SQLiteBlobFile import SQLiteBlobsFile, WriteSQLiteBlobs
from .utils import (
    CompressionError,
    IndexNotOpenError,
    InvalidIndexError,
    MismatchingIndexError,
    RatarmountError,
    find_module_version,
)
from .version import __version__


def get_sqlite_tables(connection: sqlite3.Connection):
    return [x[0] for x in connection.execute('SELECT name FROM sqlite_master WHERE type="table" OR type="view"')]


def _to_version_tuple(version: str) -> Optional[Tuple[int, int, int]]:
    versionNumbers = [re.sub('[^0-9]', '', x) for x in version.split('.')]
    if len(versionNumbers) == 3:
        return (int(versionNumbers[0]), int(versionNumbers[1]), int(versionNumbers[2]))
    return None


@dataclass
class SQLiteIndexedTarUserData:
    # fmt: off
    offset         : int
    offsetheader   : int
    istar          : bool
    issparse       : bool
    isgenerated    : bool
    recursiondepth : int
    # fmt: on


class SQLiteIndex:
    """
    This class reads once through the whole TAR archive and stores TAR file offsets
    for all contained files in an index to support fast seeking to a given file.
    """

    # Considerations for a new backwards-incompatible index format, which probably will never happen...:
    #  - Avoid redundant prefixes by storing ID-ing them. This can be slow though :(
    #    This would not be downwards-compatible because a view was ignored by my existence check in ratarmount 1.0.0.
    #  - Maybe put each recursive archive into a separate table? the table name would be some ID that could
    #    be looked up in a separate 'mountpoints' table, which maps a path to a table ID and maybe a recursion
    #    depth.
    #  - Make it possible to undo arbitrarily many outer compressions layers.
    #    This might be doable in a backwards-compatible manner already.
    #    Simply add a new table 'compressions' to store indexes for all but the outermost compression.
    #    Old versions would only be able to undo the outermost compression, while newer versions would be able
    #    to use this additional information to speed things up.
    #    It might however be difficult to implement correctly with subfolders and such, i.e., do not simply
    #    undo the compression layer with compression-backend-open calls but implement a full "SingleFileMountSource"
    #    hierarchy.
    #     -> for backward compatibility note that the innermost layer would have to be in the existing 'gzipindex'
    #        table as the others would be undone by other mount layers in older versions!
    #     -> But(!), the index would be stored for the outermost file, i.e., it would contain files and offsets,
    #        which would be loaded by older versions, without fully undoing all compressions!
    #     -> So, probably not backward compatible after all.
    #  - Make it possible to store all indexes for recursive ZIPs, Libarchive, RAR, which already support SQLiteIndex
    #    as backend, in a single database file to speed up loading of recursive archives!
    #    Would have to be some kind of multiplexing, e.g., with 'files_<ID suffix>' tables.
    #    I would only have to make the SQLiteIndex default table adjustable.
    # May be possible in a backwards-compatible (old versions will ignore it) manner:
    #  - Make it such that recursion depth changes, at least decrementing ones, do not require
    #    a full rebuild of the index. I think I can get this to work with the current version already.
    #  - Make it possible to undo inner compression layers, heck maybe even combine all inner compressions
    #    of the same type (gzip, bzip2, zstandard), create a StenciledFile so that the number of rapidgzip
    #    background instances gets reduced.
    #    - Maybe add a simple "(get/set)CompressionIndex" interface and store indexes as blobs for each file
    #      if possible, ID'd by FileInfo, i.e., by offset/offsetheader and such or maybe even the rowid.
    #
    # Version 0.1.0:
    #   - Initial version
    # Version 0.2.0:
    #   - Add sparse support and 'offsetheader' and 'issparse' columns to the SQLite database
    #   - Add TAR file size metadata in order to quickly check whether the TAR changed
    #   - Add 'offsetheader' to the primary key of the 'files' table so that files which were
    #     updated in the TAR can still be accessed if necessary.
    # Version 0.3.0:
    #   - Add arguments influencing the created index to 'metadata' table (ignore-zeros, recursive, ...)
    # Version 0.4.0:
    #   - Add 'gzipindexes' table, which may contain multiple blobs in contrast to 'gzipindex' table.
    # Version 0.5.0:
    #   - Add 'backendName' to 'metadata' table. Indexes created by different backends should by default
    #     be assumed to be incompatible, especially for chimera files, but also when one was created with
    #     libarchive, then it will not be readable with the SQLiteIndexedTar backend because it does not
    #     collect data offsets.
    #   - Add 'isGnuIncremental' to 'metadata' table.
    # Version 0.6.0:
    #   - Add 'isgenerated' and 'recursiondepth' columns.
    #     From what I can gather, the SQLite index is used in such a way that adding columns does not break anything.
    #     All SELECT statements either explicitly specify columns, or access columns via row[index] or row[columnName],
    #     e.g., in _row_to_file_info and SQLiteIndexedTar._check_rows_validity.
    #     I have also checked for all users of SQLiteIndex: TAR, ZIP, SquashFS, libarchive.
    #     I created the index with the new version and accessed it with an old ratarmount version to test compatibility.
    #     Libarchive (currently) only creates SQLite indexes in memory, ergo has no compatibility issues!
    #     Unfortunately, the _check_rows_validity will fail because the loaded vs. generated tuple lengths will differ!
    #     This is only called when 'backendName' does not exist, i.e., for indexes with version < 0.5.0 and when
    #     the TAR was detected as having been appended to. So, it's probably negligible. Honestly, who even uses
    #     an older ratarmount version with a newer index. Probably only if the index was distributed somewhere, i.e.,
    #     it can be reloaded if it was overwritten, I hope.
    #   - Add 'xattrs' table
    __version__ = '0.6.0'

    MAGIC_BYTES = b'SQLite format 3\x00'  # If it is encrypted, the first 16 B are the (random) salt.
    NUMBER_OF_METADATA_TO_VERIFY = 1000  # shouldn't take more than 1 second according to benchmarks

    # The maximum blob size configured by SQLite is exactly 1 GB, see https://www.sqlite.org/limits.html
    # Therefore, this should be smaller. Another argument for making it smaller is that this blob size
    # will be held fully in memory temporarily.
    # But, making it too small would result in too many non-backwards compatible indexes being created.
    _MAX_BLOB_SIZE = 256 * 1024 * 1024  # 256 MiB

    _CREATE_FILES_TABLE = """
        CREATE TABLE "files" (
            "path"           VARCHAR(65535) NOT NULL,  /* path with leading and without trailing slash */
            "name"           VARCHAR(65535) NOT NULL,
            "offsetheader"   INTEGER,  /* seek offset from TAR file where the TAR metadata for this file resides */
            "offset"         INTEGER,  /* seek offset from TAR file where these file's contents resides */
            "size"           INTEGER,
            "mtime"          REAL,
            "mode"           INTEGER,
            "type"           INTEGER,
            "linkname"       VARCHAR(65535),
            "uid"            INTEGER,
            "gid"            INTEGER,
            /* True for valid TAR files. Internally used to determine where to mount recursive TAR files. */
            "istar"          BOOL   ,
            "issparse"       BOOL   ,  /* For sparse files the file size refers to the expanded size! */
            "isgenerated"    BOOL   ,  /* True for entries generated for parent folders by ratarmount. */
            "recursiondepth" INTEGER,  /* Normally 0. 1+ if the file is recursively in an archive. */
            /* See SQL benchmarks for decision on the primary key.
             * See also https://www.sqlite.org/optoverview.html
             * (path,name) tuples might appear multiple times in a TAR if it got updated.
             * In order to also be able to show older versions, we need to add
             * the offsetheader column to the primary key. */
            PRIMARY KEY (path,name,offsetheader)
        );"""

    _CREATE_XATTRS_TABLE = """
        CREATE TABLE IF NOT EXISTS "xattrkeys" ( "name" VARCHAR(65535) PRIMARY KEY );

        CREATE TABLE IF NOT EXISTS "xattrsdata" (
            "offsetheader" INTEGER,
            "keyid" INTEGER,
            "value" VARCHAR(65535),  /* Binary Data (Python Bytes) */
            PRIMARY KEY (offsetheader,keyid)
        );

        CREATE VIEW IF NOT EXISTS "xattrs" ( "offsetheader", "key", "value" ) AS
            SELECT offsetheader, xattrkeys.name, value FROM "xattrsdata"
            INNER JOIN xattrkeys ON xattrkeys.rowid = xattrsdata.keyid;

        CREATE TRIGGER IF NOT EXISTS "xattrs_insert" INSTEAD OF INSERT ON "xattrs"
        BEGIN
            INSERT OR IGNORE INTO xattrkeys(name) VALUES (NEW.key);
            INSERT OR IGNORE INTO xattrsdata(offsetheader, keyid, value) VALUES(
                NEW.offsetheader,
                (SELECT xattrkeys.rowid FROM xattrkeys WHERE name = NEW.key),
                NEW.value
            );
        END;"""

    _CREATE_FILESTMP_TABLE = """
        /* "A table created using CREATE TABLE AS has no PRIMARY KEY and no constraints of any kind"
         * Therefore, it will not be sorted and inserting will be faster! */
        CREATE TABLE "filestmp" AS SELECT * FROM "files" WHERE 0;"""

    _CREATE_PARENT_FOLDERS_TABLE = """
        CREATE TABLE "parentfolders" (
            "path"          VARCHAR(65535) NOT NULL,
            "name"          VARCHAR(65535) NOT NULL,
            "offsetheader"  INTEGER,
            "offset"        INTEGER,
            PRIMARY KEY (path,name)
            UNIQUE (path,name)
        );"""

    _CREATE_METADATA_TABLE = """
        /* Empty table whose sole existence specifies that the archive has been fully processed.
         * Common keys: tarstats, arguments, isGnuIncremental, backendName */
        CREATE TABLE IF NOT EXISTS "metadata" (
            "key"      VARCHAR(65535) NOT NULL, /* e.g. "tarsize" */
            "value"    VARCHAR(65535) NOT NULL  /* e.g. size in bytes as integer */
        );
    """

    _CREATE_VERSIONS_TABLE = """
        /* This table's sole existence specifies that we finished iterating the tar for older ratarmount versions */
        CREATE TABLE IF NOT EXISTS "versions" (
            "name"     VARCHAR(65535) NOT NULL, /* which component the version belongs to */
            "version"  VARCHAR(65535) NOT NULL, /* free form version string */
            /* Semantic Versioning 2.0.0 (semver.org) parts if they can be specified:
             *   MAJOR version when you make incompatible API changes,
             *   MINOR version when you add functionality in a backwards compatible manner, and
             *   PATCH version when you make backwards compatible bug fixes. */
            "major"    INTEGER,
            "minor"    INTEGER,
            "patch"    INTEGER
        );
    """

    # Check some of the first and last files in the archive and some random selection in between.
    # Do not verify folders because parent folders and root get automatically added!
    # Ignore rows with isTar=True because recomputing this would require trying an expensive recursive mount.
    # We cannot simply ignore rows with NOT isTar because there will be multiple entries with the same
    # offset header for those recursive entries.
    # We also need to ignore all rows for recursive entries because they will have the parent TAR path prepended,
    # which the error check code is not expecting.
    # This check is done by creating a subquery/table that gathers all recursive archives (isTar==True) that
    # enclose the current rows offset. If there is no such archive, then the current row is non-recursive
    # and safe to check. Unfortunately, it seems that this is quadratic in complexity and basically hangs for
    # larger files.
    # FROM "files" AS t1 WHERE (mode & {stat.S_IFREG}) != 0 AND NOT EXISTS (
    #     SELECT 1 FROM "files" AS t2 WHERE isTar AND t1.offsetheader BETWEEN t2.offset AND t2.offset + t2.size - 1
    # )
    # As an alternative, exempt the path from the consistency check.
    # Note also that for pure compressed files such as simple.bz2, the offsetheader can be None.
    # These rows should also be filtered.
    FROM_REGULAR_FILES = f"""FROM "files" WHERE (mode & {stat.S_IFREG}) != 0 AND offsetheader IS NOT NULL"""

    def __init__(
        self,
        indexFilePath: Optional[str],
        indexFolders: Optional[List[str]] = None,
        archiveFilePath: Optional[str] = None,
        *,  # force all parameters after to be keyword-only
        encoding: str = tarfile.ENCODING,
        printDebug: int = 0,
        preferMemory: bool = False,
        indexMinimumFileCount: int = 0,
        backendName: str = '',
        ignoreCurrentFolder: bool = False,
        deleteInvalidIndexes: bool = True,
    ):
        """
        indexFilePath
            Path to the index file. This takes precedence over defaultIndexFilePath.
            If it is ':memory:', then the SQLite database will be kept in memory
            and not stored to the file system at any point.
        indexFolders
            Specify one or multiple paths for storing .index.sqlite files. Paths will be tested for
            suitability in the given order. An empty path will be interpreted as the location in which
            the archive resides in.
        preferMemory
            If True, then load existing indexes and write to explicitly given index file paths but
            if no such things are given, then create the new index in memory as if indexFilePath
            = ':memory:' was specified.
        indexMinimumFileCount
            If > 0, then open new databases in memory and write them out if this threshold has been
            exceeded. It may also be written to a file if a gzip index is stored.
        backendName
            The backend name to be stored as metadata and to determine compatibility of found indexes.
        ignoreCurrentFolder
            If true, then do not store the index into the current path. This was introduced for URL
            opened as file objects but may be useful for any archive given via a file object.
        """

        self.printDebug = printDebug
        self.sqlConnection: Optional[sqlite3.Connection] = None
        # Will hold the actually opened valid path to an index file
        self.indexFilePath: Optional[str] = None
        # This is true if the index file found was compressed or an URL and had to be downloaded
        # and/or extracted into a temporary folder.
        self.indexFilePathDeleteOnClose = False
        self.deleteInvalidIndexes = deleteInvalidIndexes
        self.encoding = encoding
        self.possibleIndexFilePaths = SQLiteIndex.get_possible_index_file_paths(
            indexFilePath,
            indexFolders,
            archiveFilePath,
            ignoreCurrentFolder,
        )
        # stores which parent folders were last tried to add to database and therefore do exist
        self.parentFolderCache: List[Tuple[str, str]] = []
        self.preferMemory = preferMemory
        self.indexMinimumFileCount = indexMinimumFileCount
        self.backendName = backendName
        self._insertedRowCount = 0
        self._temporaryIndexFile: Optional[Any] = None

        # Ignore minimum file count option if an index file path or index folder is configured.
        # For latter, ignore the special empty folder [''], which means that the indexes are stored
        # besides the archives.
        # Why all this exceptions? Because the index-minimum-file-count, which is set by default,
        # is only intended to avoid littering folders with index files in the common use case of mounting
        # a folder of archives or single small archives recursively. If the user goes through the trouble
        # of specifying an index file or folder, then littering should not be a problem as index creations
        # are expected by the user.
        # Guard this feature for Python >= 3.6 because it uses sqlite3.Connection.backup, which was only
        # introduced in Python 3.7!
        if self.indexMinimumFileCount > 0 and not indexFilePath and sys.version_info[0:2] >= (3, 7):
            if self.printDebug >= 3:
                print(
                    f"[Info] Because of the given positive index minimum file count ({self.indexMinimumFileCount}) "
                    "and because no explicit index file path is given, try to open an SQLite database in memory first."
                )
            self.preferMemory = True

    @staticmethod
    def get_possible_index_file_paths(
        indexFilePath: Optional[str],
        indexFolders: Optional[List[str]] = None,
        archiveFilePath: Optional[str] = None,
        ignoreCurrentFolder: bool = False,
    ) -> List[str]:
        if indexFilePath == ':memory:':
            return []

        possibleIndexFilePaths = []
        if indexFilePath:
            # Prior versions did simply return indexFilePath as the only possible path if it was specified.
            # This worked well enough because if the path did not exist, it would simply be created.
            # However, for non-writable locations, or if parent folders are missing, or for remote URLs,
            # this will fail badly and result in TAR files being opened with the fallback, libarchive, instead,
            # which is unwanted behavior. It should fall back to another index storage location instead.
            # Or even better, it should escalate the error to the user, but that seems too difficult with the
            # current trial-and-error architecture for opening archives.
            if '://' not in indexFilePath:
                return [os.path.abspath(os.path.expanduser(indexFilePath))]
            possibleIndexFilePaths.append(indexFilePath)

        if not archiveFilePath:
            return possibleIndexFilePaths

        # Look for default (compressed) indexes. The compressed ones should only be used as fallbacks
        # because they are less performant and because we do not want to accidentally created index files
        # with a compressed extension even though it is uncompressed. The latter reason is also why we
        # check for file existence before adding it as a default, although I think it might not be necessary. */
        defaultIndexFilePath = archiveFilePath + ".index.sqlite"
        defaultIndexFilePaths = [defaultIndexFilePath]
        for extensions in [
            FILE_FORMATS[fid].extensions for backend in COMPRESSION_BACKENDS.values() for fid in backend.formats
        ]:
            for extension in extensions:
                path = defaultIndexFilePath + '.' + extension
                if os.path.isfile(path) and os.stat(path).st_size > 0:
                    defaultIndexFilePaths.append(path)

        if not indexFolders:
            possibleIndexFilePaths.extend(defaultIndexFilePaths)
            possibleIndexFilePaths.append(':memory:')
            return possibleIndexFilePaths

        indexPathAsName = defaultIndexFilePath.replace("/", "_")
        for folder in indexFolders:
            if folder:
                indexPath = os.path.join(folder, indexPathAsName)
                possibleIndexFilePaths.append(os.path.abspath(os.path.expanduser(indexPath)))
            elif not ignoreCurrentFolder:
                possibleIndexFilePaths.extend(defaultIndexFilePaths)
        return possibleIndexFilePaths

    def clear_indexes(self):
        for indexPath in self.possibleIndexFilePaths:
            if os.path.isfile(indexPath):
                os.remove(indexPath)

    def open_existing(self, checkMetadata: Optional[Callable[[Dict[str, Any]], None]] = None, readOnly: bool = False):
        """Tries to find an already existing index."""
        for indexPath in self.possibleIndexFilePaths:
            if self._try_load_index(indexPath, checkMetadata=checkMetadata, readOnly=readOnly):
                break

    def open_in_memory(self):
        self.indexFilePath, self.sqlConnection = SQLiteIndex._open_path(':memory:')

    def open_writable(self):
        if self.possibleIndexFilePaths and not self.preferMemory:
            for indexPath in self.possibleIndexFilePaths:
                if SQLiteIndex._path_is_writable(
                    indexPath, printDebug=self.printDebug
                ) and SQLiteIndex._path_can_be_used_for_sqlite(indexPath, printDebug=self.printDebug):
                    self.indexFilePath, self.sqlConnection = SQLiteIndex._open_path(indexPath)
                    break
        else:
            if self.printDebug >= 3 and self.preferMemory:
                print("[Info] Create new index in memory because memory is to be preferred, e.g., for small archives.")
            self.indexFilePath, self.sqlConnection = SQLiteIndex._open_path(':memory:')

        if not self.index_is_loaded():
            raise InvalidIndexError(
                "Could not find any existing index or writable location for an index in "
                + str(self.possibleIndexFilePaths)
            )

    def __enter__(self):
        return self

    def __del__(self):
        # This is important in case SQLiteIndex is not used with a context manager
        # and the constructor raises an exception. There is no clean way to close it in that case!
        self.close()
        if hasattr(super(), '__del__'):
            super().__del__()

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    def close(self):
        if self.sqlConnection:
            try:
                self.sqlConnection.commit()
                self.sqlConnection.close()
            except sqlite3.ProgrammingError:
                # Ignore "Cannot operate on a closed database."
                pass
            self.sqlConnection = None

        self._set_index_file_path(None)

    def get_connection(self) -> sqlite3.Connection:
        if self.sqlConnection:
            return self.sqlConnection
        raise IndexNotOpenError("This method can not be called without an opened index database!")

    def _store_versions_metadata(self) -> None:
        connection = self.get_connection()

        try:
            connection.executescript(SQLiteIndex._CREATE_VERSIONS_TABLE)
        except Exception as exception:
            print("[Warning] There was an error when adding metadata information. Index loading might not work.")
            if self.printDebug >= 2:
                print(exception)
            if self.printDebug >= 3:
                traceback.print_exc()

        try:

            def make_version_row(
                versionName: str, version: str
            ) -> Tuple[str, str, Optional[str], Optional[str], Optional[str]]:
                versionNumbers = [re.sub('[^0-9]', '', x) for x in version.split('.')]
                return (
                    versionName,
                    version,
                    versionNumbers[0] if len(versionNumbers) > 0 else None,
                    versionNumbers[1] if len(versionNumbers) > 1 else None,
                    versionNumbers[2] if len(versionNumbers) > 2 else None,
                )

            versions = [
                make_version_row('ratarmount', __version__),
                make_version_row('index', SQLiteIndex.__version__),
            ]

            for moduleName in {
                module
                for info in COMPRESSION_BACKENDS.values()
                for module, _ in info.requiredModules
                if module in sys.modules
            }:
                moduleVersion = find_module_version(sys.modules[moduleName])
                if moduleVersion:
                    versions += [make_version_row(moduleName, moduleVersion)]

            connection.executemany('INSERT OR REPLACE INTO "versions" VALUES (?,?,?,?,?)', versions)
        except Exception as exception:
            print("[Warning] There was an error when adding version information.")
            if self.printDebug >= 2:
                print(exception)
            if self.printDebug >= 3:
                traceback.print_exc()

    def _store_file_metadata(self, filePath: AnyStr) -> None:
        """Adds some consistency meta information to recognize the need to update the cached TAR index"""
        try:
            tarStats = os.stat(filePath)
            serializedTarStats = json.dumps(
                {attr: getattr(tarStats, attr) for attr in dir(tarStats) if attr.startswith('st_')}
            )
            self.store_metadata_key_value("tarstats", serializedTarStats)
        except Exception as exception:
            print("[Warning] There was an error when adding file metadata.")
            print("[Warning] Automatic detection of changed TAR files during index loading might not work.")
            if self.printDebug >= 2:
                print(exception)
            if self.printDebug >= 3:
                traceback.print_exc()

    def store_metadata_key_value(self, key: AnyStr, value: Union[str, bytes]) -> None:
        connection = self.get_connection()
        connection.executescript(SQLiteIndex._CREATE_METADATA_TABLE)

        try:
            connection.execute('INSERT OR REPLACE INTO "metadata" VALUES (?,?)', (key, value))
        except Exception as exception:
            if self.printDebug >= 2:
                print(exception)
            print("[Warning] There was an error when adding argument metadata.")
            print("[Warning] Automatic detection of changed arguments files during index loading might not work.")

        connection.commit()

    def store_metadata(self, metadata: AnyStr, filePath: Optional[AnyStr] = None) -> None:
        self._store_versions_metadata()
        self.store_metadata_key_value('backendName', self.backendName)
        if filePath:
            self._store_file_metadata(filePath)
        self.store_metadata_key_value('arguments', metadata)
        self.store_metadata_key_value('backendName', self.backendName)

    def drop_metadata(self):
        self.get_connection().executescript(
            """
            DROP TABLE IF EXISTS metadata;
            DROP TABLE IF EXISTS versions;
            """
        )

    def try_to_open_first_file(self, openByPath):
        # Get first row that has the regular file bit set in mode (stat.S_IFREG == 32768 == 1<<15).
        result = self.get_connection().execute(
            f"""SELECT path,name {SQLiteIndex.FROM_REGULAR_FILES} ORDER BY "offsetheader" ASC LIMIT 1;"""
        )
        if not result:
            return
        firstFile = result.fetchone()
        if not firstFile:
            return

        if self.printDebug >= 2:
            print(
                "[Info] The index contains no backend name. Therefore, will try to open the first file as "
                "an integrity check."
            )
        try:
            with openByPath(firstFile[0] + '/' + firstFile[1]) as file:
                file.read(1)
        except Exception as exception:
            if self.printDebug >= 2:
                print("[Info] Trying to open the first file raised an exception:", exception)
            if self.printDebug >= 3:
                traceback.print_exc()
            raise InvalidIndexError("Integrity check of opening the first file failed.") from exception

    @staticmethod
    def check_archive_stats(
        archiveFilePath: Optional[str], metadata: Dict[str, Any], verifyModificationTime: bool
    ) -> None:
        """Raises an exception if the metadata mismatches so much that the index has to be treated as incompatible."""

        if 'tarstats' not in metadata:
            return

        if not archiveFilePath:
            raise InvalidIndexError("Archive contains file stats but cannot stat real archive!")

        storedStats = json.loads(metadata['tarstats'])
        archiveStats = os.stat(archiveFilePath)

        if (
            hasattr(archiveStats, 'st_size')
            and 'st_size' in storedStats
            and archiveStats.st_size < storedStats['st_size']
        ):
            raise InvalidIndexError(
                f"Archive for this SQLite index has shrunk in size from "
                f"{storedStats['st_size']} to {archiveStats.st_size}"
            )

        # Only happens very rarely, e.g., for more recent files with the same size.
        if (
            verifyModificationTime
            and hasattr(archiveStats, "st_mtime")
            and 'st_mtime' in storedStats
            and archiveStats.st_mtime != storedStats['st_mtime']
        ):
            raise InvalidIndexError(
                f"The modification date for the archive file {storedStats['st_mtime']} "
                f"to this SQLite index has changed ({archiveStats.st_mtime!s})",
            )

    @staticmethod
    def check_metadata_arguments(metadata: Dict, arguments, argumentsToCheck: List[str]):
        # Check arguments used to create the found index.
        # These are only warnings and not forcing a rebuild by default.
        # TODO: Add option to force index rebuild on metadata mismatch?
        differingArgs = [
            (arg, metadata[arg], getattr(arguments, arg))
            for arg in argumentsToCheck
            if arg in metadata and hasattr(arguments, arg) and metadata[arg] != getattr(arguments, arg)
        ]
        if differingArgs:
            print("[Warning] The arguments used for creating the found index differ from the arguments ")
            print("[Warning] given for mounting the archive now. In order to apply these changes, ")
            print("[Warning] recreate the index using the --recreate-index option!")
            for arg, oldState, newState in differingArgs:
                print(f"[Warning] {arg}: index: {oldState}, current: {newState}")

    def check_metadata_backend(self, metadata: Dict):
        # When opening an index without a backend via SQLiteIndexMountSource directly, it should not be checked.
        if not self.backendName:
            return

        # Because of a lack of sufficient foresight, the backend name was not added to the index in older versions.
        backendName = metadata.get('backendName')
        if isinstance(backendName, str) and backendName != self.backendName:
            raise MismatchingIndexError(
                f"Cannot open an index created with backend '{backendName}'!\n"
                f"Will stop trying to open the archive with backend '{self.backendName}'.\n"
                f"Use --recreate-index if '{backendName}' is not installed."
            )

    def get_index_version(self):
        return self.get_connection().execute("""SELECT version FROM versions WHERE name == 'index';""").fetchone()[0]

    @staticmethod
    def _path_is_writable(path: str, printDebug: int = 0) -> bool:
        # Writing indexes to remote filesystems currently not supported and we need to take care that URLs
        # are not interpreted as local file paths, i.e., creating an ftp: folder with a user:password@host subfolder.
        if '://' in path:
            return False

        try:
            folder = os.path.dirname(path)
            if folder:
                os.makedirs(folder, exist_ok=True)

            file = Path(path)
            file.write_bytes(b'\0' * 1024 * 1024)
            file.unlink()

            return True

        except PermissionError:
            if printDebug >= 3:
                print(f"Insufficient permissions to write to: {path}")

        except OSError:
            if printDebug >= 2:
                traceback.print_exc()
                print("Could not create file:", path)

        return False

    @staticmethod
    def _path_can_be_used_for_sqlite(path: str, printDebug: int = 0) -> bool:
        if not SQLiteIndex._path_is_writable(path, printDebug):
            return False

        fileExisted = os.path.isfile(path)
        try:
            folder = os.path.dirname(path)
            if folder:
                os.makedirs(folder, exist_ok=True)

            connection = SQLiteIndex._open_sql_db(path)
            connection.executescript('CREATE TABLE "files" ( "path" VARCHAR(65535) NOT NULL );')
            connection.commit()
            connection.close()
            return True
        except sqlite3.OperationalError:
            if printDebug >= 2:
                traceback.print_exc()
                print("Could not create SQLite database at:", path)
        finally:
            if not fileExisted and os.path.isfile(path):
                SQLiteIndex._unchecked_remove(path)

        return False

    @staticmethod
    def _unchecked_remove(path: Optional[AnyStr]):
        """
        Often cleanup is good manners but it would only be obnoxious if ratarmount crashed on unnecessary cleanup.
        """
        if not path or not os.path.exists(path):
            return

        try:
            os.remove(path)
        except Exception:
            print("[Warning] Could not remove:", path)

    @staticmethod
    def _open_sql_db(path: AnyStr, **kwargs) -> sqlite3.Connection:
        # Even when given a relative path, sqlite3.connect seems to access the path absolute instead of from the
        # current working directory! This will lead to hangs for self-bind mounts because of a recursive FUSE call.
        # I don't know how to circumvent this here. The caller must ensure that the possible index file paths
        # do not point to a self-bind mount point.
        sqlConnection = sqlite3.connect(path, **kwargs)
        sqlConnection.row_factory = sqlite3.Row
        sqlConnection.executescript(
            # Locking mode exclusive leads to a measurable speedup. E.g., find on 2k recursive files tar
            # improves from ~1s to ~0.4s!
            # https://blog.devart.com/increasing-sqlite-performance.html
            """
            PRAGMA LOCKING_MODE = EXCLUSIVE;
            PRAGMA TEMP_STORE = MEMORY;
            PRAGMA JOURNAL_MODE = OFF;
            PRAGMA SYNCHRONOUS = OFF;
            """
        )
        return sqlConnection

    @staticmethod
    def _open_path(indexFilePath: Optional[str], printDebug: int = 0) -> Tuple[str, sqlite3.Connection]:
        indexFilePath = indexFilePath or ':memory:'

        if printDebug >= 1:
            print("Creating new SQLite index database at", indexFilePath)

        sqlConnection = SQLiteIndex._open_sql_db(indexFilePath)
        tables = get_sqlite_tables(sqlConnection)
        if {"files", "filestmp", "parentfolders"}.intersection(tables):
            raise InvalidIndexError(
                f"The index file {indexFilePath} already seems to contain a table. Please specify --recreate-index."
            )
        sqlConnection.executescript(SQLiteIndex._CREATE_FILES_TABLE)
        sqlConnection.executescript(SQLiteIndex._CREATE_XATTRS_TABLE)

        return indexFilePath, sqlConnection

    def reload_index_read_only(self):
        if not self.indexFilePath or self.indexFilePath == ':memory:' or not self.sqlConnection:
            return

        self.sqlConnection.commit()
        self.sqlConnection.close()

        uriPath = urllib.parse.quote(self.indexFilePath)
        # check_same_thread=False can be used because it is read-only anyway and it allows to enable FUSE multithreading
        self.sqlConnection = SQLiteIndex._open_sql_db(f"file:{uriPath}?mode=ro", uri=True, check_same_thread=False)

    def _reload_index_on_disk(self):
        if self.printDebug >= 2:
            print("[Info] Try to reopen SQLite database on disk at:", self.indexFilePath)
            print("other index paths:", self.possibleIndexFilePaths)
        if not self.indexFilePath or self.indexFilePath != ':memory:' or not self.sqlConnection:
            return

        oldIndexFilePath, oldSqlConnection = self.indexFilePath, self.sqlConnection
        self.preferMemory = False
        self.open_writable()
        if oldIndexFilePath == self.indexFilePath:
            if self.printDebug >= 3:
                print(f"[Info] Tried to write the database to disk but found no other path than: {self.indexFilePath}")
            self.sqlConnection.close()
            self.indexFilePath, self.sqlConnection = oldIndexFilePath, oldSqlConnection
            return

        if self.printDebug >= 3:
            print(f"[Info] Back up database from {oldIndexFilePath} -> {self.indexFilePath}")
        oldSqlConnection.commit()
        oldSqlConnection.backup(self.sqlConnection)
        oldSqlConnection.close()

    def ensure_intermediary_tables(self):
        connection = self.get_connection()
        tables = get_sqlite_tables(connection)

        if ("filestmp" in tables) != ("parentfolders" in tables):
            raise InvalidIndexError(
                "The index file is in an invalid state because it contains some tables and misses others. "
                "Please specify --recreate-index to overwrite the existing index."
            )

        if "filestmp" not in tables and "parentfolders" not in tables:
            connection.execute(SQLiteIndex._CREATE_FILESTMP_TABLE)
            connection.execute(SQLiteIndex._CREATE_PARENT_FOLDERS_TABLE)

    def finalize(self):
        try:
            # Note that the with-statement does not automatically close a connection!
            # https://discuss.python.org/t/implicitly-close-sqlite3-connections-with-context-managers/33320
            connection = sqlite3.connect(":memory:")
            queriedLibSqliteVersion = connection.execute("select sqlite_version();").fetchone()
            connection.close()
            libSqliteVersion = tuple(int(x) for x in queriedLibSqliteVersion[0].split('.'))
        except Exception:
            libSqliteVersion = (0, 0, 0)

        searchByTuple = """(path,name) NOT IN ( SELECT path,name"""
        searchByConcat = """path || "/" || name NOT IN ( SELECT path || "/" || name"""

        cleanUpDatabase = f"""
            INSERT OR REPLACE INTO "files" SELECT * FROM "filestmp" ORDER BY "path","name",rowid;
            DROP TABLE "filestmp";
            INSERT OR IGNORE INTO "files"
                /* path name offsetheader offset size mtime mode type */
                SELECT path,name,offsetheader,offset,0,0,{int(0o555 | stat.S_IFDIR)},{int(tarfile.DIRTYPE)},
                       /* linkname uid gid istar issparse isgenerated recursiondepth */
                       "",0,0,0,0,0,0
                FROM "parentfolders"
                WHERE {searchByTuple if libSqliteVersion >= (3, 22, 0) else searchByConcat}
                    FROM "files" WHERE mode & (1 << 14) != 0
                )
                ORDER BY "path","name";
            DROP TABLE "parentfolders";
            VACUUM;  /* According to benchmarks, this is important to reduce the size after the table drop! */
            PRAGMA optimize;
        """

        self.get_connection().executescript(cleanUpDatabase)

    def file_count(self) -> int:
        return self.get_connection().execute('SELECT COUNT(*) FROM "files";').fetchone()[0]

    @staticmethod
    def _row_to_file_info(row: Dict[str, Any]) -> FileInfo:
        # fmt: off
        userData = SQLiteIndexedTarUserData(
            offset         = row['offset'],
            offsetheader   = row['offsetheader'] if 'offsetheader' in row.keys() else 0,  # noqa: SIM118
            istar          = row['istar'],
            issparse       = row['issparse'] if 'issparse' in row.keys() else False,  # noqa: SIM118
            isgenerated    = row['isgenerated'] if 'isgenerated' in row.keys() else False,  # noqa: SIM118
            recursiondepth = row['recursiondepth'] if 'recursiondepth' in row.keys() else False,  # noqa: SIM118
        )
        return FileInfo(
            size     = row['size'],
            mtime    = row['mtime'],
            mode     = row['mode'],
            linkname = row['linkname'],
            uid      = row['uid'],
            gid      = row['gid'],
            userdata = [userData],
        )
        # fmt: on

    @staticmethod
    def normpath(path: str):
        # Add a leading '/' as a convention where '/' represents the TAR root folder.
        # Also strips trailing '/' except for a single '/'.
        # Partly, done because fusepy specifies paths in a mounted directory like this
        # os.normpath does not delete duplicate '/' at beginning of string!
        # os.path.normpath can remove suffixed folder/./ path specifications but it can't remove
        # a leading dot that's why we prefix a leading slash also before calling normpath.
        return '/' + os.path.normpath('/' + path).lstrip('/')

    @staticmethod
    def _query_normpath(path: str):
        # os.path.normpath also collapses /../ into / and, because we prepend /, ../ gets collapsed to /.
        # Note that normpath does not collapse leading double slash, but all other number of leading slashes!
        # This effect is good to have for inserting rows but not for querying rows.
        return '/' + os.path.normpath(path if path.startswith('../') else '/' + path).lstrip('/')

    def list(self, path: str) -> Optional[Dict[str, FileInfo]]:
        """
        Return a dictionary for the given directory path: { fileName : FileInfo, ... } or None
        if the path does not exist.

        There is no file version argument because it is hard to apply correctl.y
        Even if a folder was overwritten by a file, which is already not well supported by tar,
        then list for that path will still list all contents of the overwritten folder or folders,
        no matter the specified version. The file system layer has to take care that a directory
        listing is not even requested in the first place if it is not a directory.
        FUSE already does this by calling getattr for all parent folders in the specified path first.

        path : full path to file where '/' denotes TAR's root, e.g., '/', or '/foo'
        """

        # For listing directory entries the file version can't be applied meaningfully at this abstraction layer.
        # E.g., should it affect the file version of the directory to list, or should it work on the listed files
        # instead and if so how exactly if there aren't the same versions for all files available, ...?
        # Or, are folders assumed to be overwritten by a new folder entry in a TAR or should they be union mounted?
        # If they should be union mounted, like is the case now, then the folder version only makes sense for
        # its attributes.
        #
        # Order by offsetheader in order to preserve the order they appear in the archive to potentially enable
        # faster access to the whoole archive when iterating over all files in order.
        # Note that Python's dictionary preserves the insertion order since Python 3.6.
        # https://docs.python.org/3.6/whatsnew/3.6.html#new-dict-implementation
        # While it was not a guarantee to stay that way, it is guaranteed for Python 3.7+:
        # > the insertion-order preservation nature of dict objects has been declared to be an official part of the
        # > Python language spec.
        # https://docs.python.org/3.11/library/stdtypes.html#dict.values
        # > Dictionaries preserve insertion order. Note that updating a key does not affect the order.
        # > Keys added after deletion are inserted at the end.

        # https://www.sqlite.org/pragma.html#pragma_table_info
        columns = [row[0] for row in self.get_connection().execute("SELECT name FROM PRAGMA_TABLE_INFO('files');")]
        selected_columns = ['name', 'size', 'mtime', 'mode', 'linkname', 'uid', 'gid', 'offset', 'istar']
        # Both were added in index 0.2.0.
        if 'offsetheader' in columns and 'issparse' in columns:
            selected_columns += ['offsetheader', 'issparse']
        # Added in index 0.6.0.
        if 'isgenerated' in columns:
            selected_columns += ['isgenerated']
        if 'recursiondepth' in columns:
            selected_columns += ['recursiondepth']

        def row_to_file_info(cursor, row) -> Tuple[str, FileInfo]:  # pylint: disable=unused-argument
            # fmt: off
            return row[0], FileInfo(
                size     = row[1],
                mtime    = row[2],
                mode     = row[3],
                linkname = row[4],
                uid      = row[5],
                gid      = row[6],
                userdata = [SQLiteIndexedTarUserData(
                    offset         = row[7],
                    offsetheader   = row[9] if len(row) > 9 else 0,
                    istar          = row[8],
                    issparse       = row[10] if len(row) > 10 else False,
                    isgenerated    = row[11] if len(row) > 11 else False,
                    recursiondepth = row[12] if len(row) > 12 else False,
                )],
            )
            # fmt: on

        oldRowFactory = self.get_connection().row_factory
        self.get_connection().row_factory = row_to_file_info
        directory: Dict[str, FileInfo] = dict(
            self.get_connection().execute(
                'SELECT ' + ','.join(selected_columns) + ' FROM "files" WHERE "path" == (?) ORDER BY "offsetheader"',
                (self._query_normpath(path).rstrip('/'),),
            )
        )
        self.get_connection().row_factory = oldRowFactory

        gotResults = bool(directory)
        directory.pop('', None)
        return directory if gotResults else None

    def list_mode(self, path: str) -> Optional[Dict[str, int]]:
        """
        Return a dictionary mapping file names to file modes for the given directory path or None
        if the path does not exist.

        path : full path to file where '/' denotes TAR's root, e.g., '/', or '/foo'
        """

        # See comments in list.
        # The only difference here is that we do not request all columns, but only two in a tuple.
        oldRowFactory = self.get_connection().row_factory
        self.get_connection().row_factory = None
        rows = self.get_connection().execute(
            'SELECT name,mode FROM "files" WHERE "path" == (?) ORDER BY "offsetheader"',
            (self._query_normpath(path).rstrip('/'),),
        )
        self.get_connection().row_factory = oldRowFactory

        directory = {row[0]: row[1] for row in rows}
        gotResults = bool(directory)
        directory.pop('', None)
        return directory if gotResults else None

    def versions(self, path: str) -> Dict[str, FileInfo]:
        """
        Return metadata for all versions of a file possibly appearing more than once
        in the index as a directory dictionary or an empty dictionary if the path does not exist.

        path : full path to file where '/' denotes TAR's root, e.g., '/', or '/foo'
        """

        path = self._query_normpath(path)
        if path == '/':
            return {'/': create_root_file_info(userdata=[SQLiteIndexedTarUserData(0, 0, False, False, True, 0)])}

        path, name = path.rsplit('/', 1)
        rows = self.get_connection().execute(
            'SELECT * FROM "files" WHERE "path" == (?) AND "name" == (?) ORDER BY "offsetheader" ASC', (path, name)
        )
        return {str(version + 1): self._row_to_file_info(row) for version, row in enumerate(rows)}

    def lookup(self, path: str, fileVersion: int = 0) -> Optional[FileInfo]:
        """
        Return the FileInfo to given path (directory or file) or None if the path does not exist.

        fileVersion : If the TAR contains the same file path multiple times, by default only the last one is shown.
                      But with this argument other versions can be queried. Version 1 is the oldest one.
                      Version 0 translates to the most recent one for compatibility with tar --occurrence=<number>.
                      Version -1 translates to the second most recent, and so on.
        """

        if not isinstance(fileVersion, int):
            raise RatarmountError("The specified file version must be an integer!")

        path = self._query_normpath(path)
        if path == '/':
            return create_root_file_info(userdata=[SQLiteIndexedTarUserData(0, 0, False, False, True, 0)])

        path, name = path.rsplit('/', 1)
        row = (
            self.get_connection()
            .execute(
                f"""
            SELECT * FROM "files"
            WHERE "path" == (?) AND "name" == (?)
            ORDER BY "offsetheader" {'DESC' if fileVersion is None or fileVersion <= 0 else 'ASC'}
            LIMIT 1 OFFSET (?);
            """,
                (path, name, 0 if fileVersion is None else fileVersion - 1 if fileVersion > 0 else -fileVersion),
            )
            .fetchone()
        )
        return self._row_to_file_info(row) if row else None

    def setxattrs(self, rows: List[Tuple[int, str, bytes]]):
        self.get_connection().executemany('INSERT OR REPLACE INTO "xattrs" VALUES (?,?,?)', rows)

    def list_xattr(self, fileInfo: FileInfo) -> List[str]:
        if not fileInfo.userdata:
            return []
        userData = fileInfo.userdata[-1]
        assert isinstance(userData, SQLiteIndexedTarUserData)

        if userData.isgenerated:
            return []

        try:
            # Look up by offsetheader seems to be unique enough for all implementations,
            # but currently only TAR supports attributes anyways.
            return [
                row[0]
                for row in self.get_connection().execute(
                    "SELECT key FROM xattrs WHERE offsetheader=(?);", (userData.offsetheader,)
                )
            ]
        except sqlite3.OperationalError:
            # May happen when loading old indexes that do not have the xattrs table.
            pass
        return []

    def get_xattr(self, fileInfo: FileInfo, key: str) -> Optional[bytes]:
        if not fileInfo.userdata:
            return None
        userData = fileInfo.userdata[-1]
        assert isinstance(userData, SQLiteIndexedTarUserData)

        if userData.isgenerated:
            return None

        try:
            row = (
                self.get_connection()
                .execute("SELECT value FROM xattrs WHERE (offsetheader,key)=(?,?);", (userData.offsetheader, key))
                .fetchone()
            )
            return row[0] if row else None
        except sqlite3.OperationalError:
            # May happen when loading old indexes that do not have the xattrs table.
            pass
        return None

    def _try_add_parent_folders(self, path: str, offsetheader: int, offset: int) -> None:
        # Add parent folders if they do not exist.
        # E.g.: path = '/a/b/c' -> paths = [('', 'a'), ('/a', 'b'), ('/a/b', 'c')]
        # Without the parentFolderCache, the additional INSERT statements increase the creation time
        # from 8.5s to 12s, so almost 50% slowdown for the 8MiB test TAR!
        pathParts = path.split("/")
        # fmt: off
        paths = [
            p
            for p in (
                ( "/".join( pathParts[:i] ), pathParts[i] )
                for i in range( 1, len( pathParts ) )
            )
            if p not in self.parentFolderCache
        ]
        # fmt: on
        if not paths:
            return

        self.parentFolderCache += paths
        # Assuming files in the TAR are sorted by hierarchy, the maximum parent folder cache size
        # gives the maximum cacheable file nesting depth. High numbers lead to higher memory usage and lookup times.
        if len(self.parentFolderCache) > 16:
            self.parentFolderCache = self.parentFolderCache[-8:]

        # TODO This method is still not perfect but I do not know how to perfect it without losing significant
        #      performance. Currently, adding implicit folders will fail when a file is overwritten implicitly with
        #      a folder and then overwritten by a file and then again overwritten by a folder. Because the parent
        #      folder was already added implicitly the first time, the second time will be skipped.
        #      To solve this, I would have to add all parent folders for all files, which might easily explode
        #      the temporary database and the indexing performance by the folder depth.
        #      Also, I do not want to add versions for a parent folder for each implicitly added parent folder for
        #      each file, so I would have to sort out those in a post-processing step. E.g., sort by offsetheader
        #      and then clean out successive implicitly added folders as long as there is no file of the same name
        #      in between.
        #      The unmentioned alternative would be to lookup paths with LIKE but that is just madness because it
        #      will have a worse complexity of O(N) instead of O(log(N)).
        self.get_connection().executemany(
            'INSERT OR IGNORE INTO "parentfolders" VALUES (?,?,?,?)',
            [(p[0], p[1], offsetheader, offset) for p in paths],
        )

    def set_file_infos(self, rows: List[Tuple]) -> None:
        if not rows:
            return

        self._insertedRowCount += len(rows)
        if (
            self._insertedRowCount > self.indexMinimumFileCount
            and self.indexFilePath == ':memory:'
            and self.preferMemory
        ):
            if self.printDebug >= 3:
                print(
                    f"[Info] Exceeded file count threshold ({self._insertedRowCount} > {self.indexMinimumFileCount}) "
                    "for metadata held in memory. Will try to reopen the database on disk."
                )
            self.preferMemory = False
            self._reload_index_on_disk()

        try:
            self.get_connection().executemany(
                'INSERT OR REPLACE INTO "files" VALUES (' + ','.join('?' * len(rows[0])) + ');', rows
            )
        except UnicodeEncodeError:
            # Fall back to separately inserting each row to find those in need of string cleaning.
            for row in rows:
                self.set_file_info(row)
            return

        for row in rows:
            self._try_add_parent_folders(row[0], row[2], row[3])

    @staticmethod
    def _escape_invalid_characters(toEscape: str, encoding: str):
        try:
            toEscape.encode()
            return toEscape
        except UnicodeEncodeError:
            return toEscape.encode(encoding, 'surrogateescape').decode(encoding, 'backslashreplace')

    def set_file_info(self, row: tuple) -> None:
        connection = self.get_connection()

        try:
            connection.execute('INSERT OR REPLACE INTO "files" VALUES (' + ','.join('?' * len(row)) + ');', row)
        except UnicodeEncodeError:
            print("[Warning] Problem caused by file name encoding when trying to insert this row:", row)
            print("[Warning] The file name will now be stored with the bad character being escaped")
            print("[Warning] instead of being correctly interpreted.")
            print("[Warning] Please specify a suitable file name encoding using, e.g., --encoding iso-8859-1!")
            print("[Warning] A list of possible encodings can be found here:")
            print("[Warning] https://docs.python.org/3/library/codecs.html#standard-encodings")

            checkedRow = []
            for x in list(row):  # check strings
                if isinstance(x, str):
                    checkedRow += [self._escape_invalid_characters(x, self.encoding)]
                else:
                    checkedRow += [x]

            connection.execute(
                'INSERT OR REPLACE INTO "files" VALUES (' + ','.join('?' * len(row)) + ');', tuple(checkedRow)
            )
            print("[Warning] The escaped inserted row is now:", row)
            print()

            self._try_add_parent_folders(self._escape_invalid_characters(row[0], self.encoding), row[2], row[3])
            return

        self._try_add_parent_folders(row[0], row[2], row[3])

    def index_is_loaded(self) -> bool:
        """Returns true if the SQLite database has been opened for reading and a "files" table exists."""
        if not self.sqlConnection:
            return False

        try:
            self.sqlConnection.execute('SELECT * FROM "files" WHERE 0 == 1;')
        except sqlite3.OperationalError:
            self.sqlConnection = None
            return False

        return True

    def _set_index_file_path(self, indexFilePath: Optional[str], deleteOnClose: bool = False):
        # This is called from __del__, so we need to account for this being called when something
        # in the constructor raises an exception and not all members of self exist.
        if (
            getattr(self, 'indexFilePath', None)
            and getattr(self, 'indexFilePathDeleteOnClose', False)
            and self.indexFilePath
            and os.path.isfile(self.indexFilePath)
        ):
            try:
                os.remove(self.indexFilePath)
            except Exception as exception:
                if self.printDebug >= 1:
                    print(
                        "[Warning] Failed to remove temporarily downloaded and/or extracted index file at:",
                        self.indexFilePath,
                        "because of:",
                        exception,
                    )

        if hasattr(self, 'indexFilePath') and hasattr(self, 'indexFilePathDeleteOnClose'):
            self.indexFilePath = indexFilePath
            self.indexFilePathDeleteOnClose = deleteOnClose

    def _load_index(
        self, indexFilePath: str, checkMetadata: Optional[Callable[[Dict[str, Any]], None]], readOnly: bool = False
    ) -> None:
        """
        Loads the given index SQLite database and checks it for validity raising an exception if it is invalid.

        checkMetadata
            A verifying callback that is called when opening an existing index. It is given the
            the dictionary of metadata in the index and should thrown an exception when the index
            should not be used, e.g., because the version is incompatible.
        """
        if self.index_is_loaded():
            return

        # Download and/or extract the file to a temporary file if necessary.

        # Strip a single file:// prefix, not any more because URL chaining is supported by fsspec,
        # to avoid useless copies to the temporary directory.
        if indexFilePath.count('://') == 1:
            fileURLPrefix = 'file://'
            if indexFilePath.startswith(fileURLPrefix):
                indexFilePath = indexFilePath[len(fileURLPrefix) :]

        temporaryFolder = os.environ.get("RATARMOUNT_INDEX_TMPDIR", None)

        def _undo_compression(file):
            compression = detect_compression(file, printDebug=self.printDebug)
            if not compression or not any(
                (compression in backend.formats) for backend in COMPRESSION_BACKENDS.values()
            ):
                return None

            if self.printDebug >= 2:
                print(f"[Info] Detected {compression}-compressed index.")

            backend = find_available_backend(compression)
            if not backend:
                packages = [
                    x[1]
                    for info in COMPRESSION_BACKENDS.values()
                    for x in info.requiredModules
                    if compression in info.formats
                ]
                raise CompressionError(
                    f"Cannot open a {compression} compressed index file {indexFilePath} "
                    f"without any of these packages: {packages}"
                )

            return backend.open(file)

        def _copy_to_temp(file):
            self._temporaryIndexFile = tempfile.NamedTemporaryFile(suffix=".tmp.sqlite.index", dir=temporaryFolder)
            # TODO add progress bar / output?
            with open(self._temporaryIndexFile.name, 'wb') as targetFile:
                shutil.copyfileobj(file, targetFile)

        if '://' in indexFilePath:
            if fsspec is None:
                raise RatarmountError(
                    "Detected an URL for the index path but fsspec could not be imported!\n"
                    "Try installing it with 'pip install fsspec' or 'pip install ratarmount[full]'."
                )

            with fsspec.open(indexFilePath) as file:
                decompressedFile = _undo_compression(file)
                with decompressedFile or file as fileToCopy:
                    _copy_to_temp(fileToCopy)
        else:
            if not os.path.isfile(indexFilePath):
                return

            with open(indexFilePath, 'rb') as file:
                decompressedFile = _undo_compression(file)
                if decompressedFile:
                    with decompressedFile:
                        _copy_to_temp(decompressedFile)
                else:
                    temporaryIndexFilePath = indexFilePath

        temporaryIndexFilePath = self._temporaryIndexFile.name if self._temporaryIndexFile else indexFilePath

        # Done downloading and/or extracting the SQLite index.

        if readOnly:
            uriPath = urllib.parse.quote(temporaryIndexFilePath)
            # check_same_thread=False can be used because it is read-only and it allows to enable FUSE multithreading
            self.sqlConnection = SQLiteIndex._open_sql_db(f"file:{uriPath}?mode=ro", uri=True, check_same_thread=False)
        else:
            self.sqlConnection = SQLiteIndex._open_sql_db(temporaryIndexFilePath)
        tables = get_sqlite_tables(self.sqlConnection)
        versions = None
        try:
            rows = self.sqlConnection.execute('SELECT * FROM versions;')
            versions = {}
            for row in rows:
                versions[row[0]] = (row[2], row[3], row[4])
        except sqlite3.OperationalError:
            pass

        try:
            # Check indexes created with bugged bz2 decoder (bug existed when I did not store versions yet)
            if 'bzip2blocks' in tables and 'versions' not in tables:
                raise InvalidIndexError(
                    "The indexes created with version 0.3.0 through 0.3.3 for bzip2 compressed archives "
                    "are very likely to be wrong because of a bzip2 decoder bug.\n"
                    "Please delete the index or call ratarmount with the --recreate-index option!"
                )

            # Check for empty or incomplete indexes. Pretty safe to rebuild the index for these as they
            # are so invalid, no one should miss them. So, recreate index by default for these cases.
            if 'files' not in tables:
                raise InvalidIndexError("SQLite index is empty")

            if 'filestmp' in tables or 'parentfolders' in tables:
                raise InvalidIndexError("SQLite index is incomplete")

            # Check for pre-sparse support indexes
            if (
                'versions' not in tables
                or 'index' not in versions
                or len(versions['index']) < 2
                or versions['index'][1] < 2
            ):
                print("[Warning] The found outdated index does not contain any sparse file information.")
                print("[Warning] The index will also miss data about multiple versions of a file.")
                print("[Warning] Please recreate the index if you have problems with those.")

            if 'metadata' in tables:
                metadata = dict(self.sqlConnection.execute('SELECT * FROM metadata;'))
                if checkMetadata:
                    self.check_metadata_backend(metadata)
                    checkMetadata(metadata)

        except Exception as e:
            # index_is_loaded checks self.sqlConnection, so close it before returning because it was found to be faulty
            with contextlib.suppress(sqlite3.Error):
                self.sqlConnection.close()
            self.sqlConnection = None

            raise e

        if self.index_is_loaded() and self.sqlConnection:
            try:
                indexVersion = self.sqlConnection.execute(
                    "SELECT major,minor FROM versions WHERE name == 'index';"
                ).fetchone()

                if indexVersion:
                    indexVersionTuple = _to_version_tuple(indexVersion)
                    indexAPIVersionTuple = _to_version_tuple(SQLiteIndex.__version__)
                    if indexVersionTuple and indexAPIVersionTuple and indexVersionTuple > indexAPIVersionTuple:
                        print("[Warning] The loaded index was created with a newer version of ratarmount.")
                        print("[Warning] If there are any problems, please update ratarmount or recreate the index")
                        print("[Warning] with this ratarmount version using the --recreate-index option!")
            except Exception:
                pass

        if self.printDebug >= 1:
            message = "Successfully loaded offset dictionary from " + str(indexFilePath)
            if temporaryIndexFilePath != indexFilePath:
                message += " temporarily downloaded/decompressed into: " + str(temporaryIndexFilePath)
            print(message)

        self._set_index_file_path(temporaryIndexFilePath)

    def _try_load_index(
        self,
        indexFilePath: str,
        checkMetadata: Optional[Callable[[Dict[str, Any]], None]] = None,
        readOnly: bool = False,
    ) -> bool:
        """Calls loadIndex if index is not loaded already and provides extensive error handling."""

        if self.index_is_loaded():
            return True

        try:
            self._load_index(indexFilePath, checkMetadata=checkMetadata, readOnly=readOnly)
        except MismatchingIndexError as e:
            raise e
        except Exception as exception:
            if self.printDebug >= 3:
                traceback.print_exc()

            print("[Warning] Could not load file:", indexFilePath)
            print("[Info] Exception:", exception)
            print("[Info] Some likely reasons for not being able to load the index file:")
            print("[Info]   - The index file has incorrect read permissions")
            print("[Info]   - The index file is incomplete because ratarmount was killed during index creation")
            print("[Info]   - The index file was detected to contain errors because of known bugs of older versions")
            print("[Info]   - The index file got corrupted because of:")
            print("[Info]     - The program exited while it was still writing the index because of:")
            print("[Info]       - the user sent SIGINT to force the program to quit")
            print("[Info]       - an internal error occurred while writing the index")
            print("[Info]       - the disk filled up while writing the index")
            print("[Info]     - Rare lowlevel corruptions caused by hardware failure")

            print("[Info] This might force a time-costly index recreation, so if it happens often")
            print("       and mounting is slow, try to find out why loading fails repeatedly,")
            print("       e.g., by opening an issue on the public github page.")

            try:
                if self.deleteInvalidIndexes and '://' not in indexFilePath:
                    os.remove(indexFilePath)
            except OSError:
                print("[Warning] Failed to remove corrupted old cached index file:", indexFilePath)

        return self.index_is_loaded()

    def clear_compression_offsets(self):
        for table in ['bzip2blocks', 'gzipindex', 'gzipindexes', 'zstdblocks']:
            self.get_connection().execute(f"DROP TABLE IF EXISTS {table}")

    def synchronize_compression_offsets(self, fileObject: IO[bytes], compression: FileFormatID):
        """
        Will load block offsets from SQLite database to backend if a fitting table exists.
        Else it will force creation and store the block offsets of the compression backend into a new table.
        """
        if compression and (not self.indexFilePath or self.indexFilePath == ':memory:'):
            if self.indexMinimumFileCount != 0 and self.printDebug >= 2:
                print(
                    f"[Info] Will try to reopen the database on disk even though the file count threshold "
                    f"({self.indexMinimumFileCount}) might not be exceeded ({self._insertedRowCount}) because "
                    f"the archive is compressed."
                )
            self._reload_index_on_disk()

        if not self.indexFilePath or self.indexFilePath == ':memory:':
            if self.printDebug >= 2:
                print("[Info] Will skip storing compression seek data because the database is in memory.")
                print("[Info] If the database is in memory, then this data will not be read anyway.")
            return

        # This should be called after the TAR file index is complete (loaded or created).
        # If the TAR file index was created, then tarfile has iterated over the whole file once
        # and therefore completed the implicit compression offset creation.
        db = self.get_connection()

        if compression in [FileFormatID.BZIP2, FileFormatID.ZSTANDARD]:
            setBlockOffsets = getattr(fileObject, 'set_block_offsets', None)
            getBlockOffsets = getattr(fileObject, 'block_offsets', None)
            if not setBlockOffsets or not getBlockOffsets:
                print("[Warning] The given file object misses the expected methods for getting/setting")
                print("[Warning] the block offsets. Subsequent loads might be slow.")
                return

            table_name = ''
            if compression == FileFormatID.BZIP2:
                table_name = 'bzip2blocks'
            elif compression == FileFormatID.ZSTANDARD:
                table_name = 'zstdblocks'

            tables = get_sqlite_tables(db)
            if table_name in tables:
                try:
                    offsets = dict(db.execute(f"SELECT blockoffset,dataoffset FROM {table_name};"))
                    setBlockOffsets(offsets)
                    return
                except Exception as exception:
                    if self.printDebug >= 2:
                        print(
                            f"[Info] Could not load {compression.name} block offset data. Will create it from scratch."
                        )
                        print(exception)
                    if self.printDebug >= 3:
                        traceback.print_exc()
            else:
                print(f"[Info] The index does not yet contain {compression.name} block offset data. Will write it out.")

            tables = get_sqlite_tables(db)
            if table_name in tables:
                db.execute(f"DROP TABLE {table_name}")
            db.execute(f"CREATE TABLE {table_name} ( blockoffset INTEGER PRIMARY KEY, dataoffset INTEGER )")
            db.executemany(f"INSERT INTO {table_name} VALUES (?,?)", getBlockOffsets().items())
            db.commit()
            return

        if (
            hasattr(fileObject, 'import_index')
            and hasattr(fileObject, 'export_index')
            and compression in [FileFormatID.GZIP, FileFormatID.ZLIB]
        ):
            tables = get_sqlite_tables(db)

            if 'gzipindex' in tables or 'gzipindexes' in tables:
                if self._load_gzip_index(fileObject, 'gzipindexes' if 'gzipindexes' in tables else 'gzipindex'):
                    return

                if self.printDebug >= 2:
                    print("[Info] Could not load gzip block offset data. Will create it from scratch.")
            else:
                if self.printDebug >= 2:
                    print("[Info] The index does not yet contain gzip block offset data. Will write it out.")

            self._store_gzip_index(fileObject)
            return

        # Note that for xz seeking, loading and storing block indexes is unnecessary because it has an index included!
        if compression == FileFormatID.XZ:
            return

        raise NotImplementedError(
            f"Could not load or store block offsets for {compression} probably because adding support was forgotten!"
        )

    def _load_gzip_index(self, fileObject: IO[bytes], table: str) -> bool:
        importIndex = getattr(fileObject, 'import_index', None)
        if not importIndex:
            return False

        connection = self.get_connection()
        try:
            t0 = time.time()
            fileobj = SQLiteBlobsFile(connection, table, 'data', buffer_size=SQLiteIndex._MAX_BLOB_SIZE)
            if 'rapidgzip' in sys.modules and isinstance(fileObject, rapidgzip.RapidgzipFile):
                importIndex(fileobj)
            else:
                # indexed_gzip 1.5.0 added support for pure Python file objects as arguments for the index!
                importIndex(fileobj=fileobj)

            # SQLiteBlobFile is rather slow to get parts of a large blob by using substr.
            # Here are some timings for 4x256 MiB blobs:
            #   buffer size / MiB | time / s
            #                 512 | 1.94 1.94 1.90 1.92 1.95
            #                 256 | 1.94 1.98 1.92 2.02 1.91
            #                 128 | 2.44 2.37 2.38 2.44 2.47
            #                  64 | 3.51 3.44 3.47 3.47 3.42
            #                  32 | 5.66 5.71 5.62 5.57 5.61
            #                  16 | 10.22 9.88 9.73 9.75 9.91
            # Writing out blobs to file / s         : 9.40 8.71 8.49 10.60 9.13
            # Importing block offsets from file / s : 0.47 0.46 0.47 0.46 0.41
            #   => With proper buffer sizes, avoiding writing out the block offsets can be 5x faster!
            # It seems to me like substr on blobs does not actually support true seeking :/
            # The blob is probably always loaded fully into memory and only then is the substring being
            # calculated. For C, there actually is an incremental blob reading interface but not for Python:
            #   https://www.sqlite.org/c3ref/blob_open.html
            #   https://bugs.python.org/issue24905
            if self.printDebug >= 3:
                print(f"[Info] Loading gzip block offsets took {time.time() - t0:.2f}s")

            return True

        except Exception as exception:
            if self.printDebug >= 1:
                print(
                    "[Warning] Encountered exception when trying to load gzip block offsets from database",
                    exception,
                )
            if self.printDebug >= 3:
                traceback.print_exc()

        return False

    def _store_gzip_index(self, fileObject: IO[bytes]):
        exportIndex = getattr(fileObject, 'export_index', None)
        if not exportIndex:
            print("[Warning] The given file object misses the expected methods for getting/setting")
            print("[Warning] the block offsets. Subsequent loads might be slow.")
            return

        # Transparently force index to be built if not already done so. build_full_index was buggy for me.
        # Seeking from end not supported, so we have to read the whole data in in a loop
        while fileObject.read(1024 * 1024):
            pass

        # Old implementation using a temporary file before copying parts of it into blobs.
        # Tested on 1 GiB of gzip block offset data (32 GiB file with block seek point spacing 1 MiB)
        #   Time / s: 23.251 22.251 23.697 23.230 22.484
        # Timings when using WriteSQLiteBlobs to write directly into the SQLite database.
        #   Time / s: 13.029 14.884 14.110 14.229 13.807

        db = self.get_connection()
        db.execute('DROP TABLE IF EXISTS "gzipindexes"')
        db.execute('CREATE TABLE gzipindexes ( data BLOB )')

        try:
            with WriteSQLiteBlobs(db, 'gzipindexes', blob_size=SQLiteIndex._MAX_BLOB_SIZE) as gzindex:
                if 'rapidgzip' in sys.modules and isinstance(fileObject, rapidgzip.RapidgzipFile):
                    # See the following link for the exception mapping done by Cython:
                    # https://cython.readthedocs.io/en/latest/src/userguide/wrapping_CPlusPlus.html#exceptions
                    exportIndex(gzindex)
                else:
                    exportIndex(fileobj=gzindex)
        except (indexed_gzip.ZranError, RuntimeError, ValueError) as exception:
            db.execute('DROP TABLE IF EXISTS "gzipindexes"')

            print("[Warning] The GZip index required for seeking could not be written to the database!")
            print("[Info] This might happen when you are out of space in your temporary file and at the")
            print("[Info] the index file location. The gzipindex size takes roughly 32kiB per 4MiB of")
            print("[Info] uncompressed(!) bytes (0.8% of the uncompressed data) by default.")

            raise RatarmountError("Could not wrote out the gzip seek database.") from exception

        blobCount = db.execute('SELECT COUNT(*) FROM gzipindexes;').fetchone()[0]
        if blobCount == 0:
            if self.printDebug >= 2:
                print("[Warning] Did not write out any gzip seek data. This should only happen if the gzip ")
                print("[Warning] size is smaller than the gzip seek point spacing.")
        elif blobCount == 1:
            # For downwards compatibility
            db.execute('DROP TABLE IF EXISTS "gzipindex";')
            db.execute('ALTER TABLE gzipindexes RENAME TO gzipindex;')

        db.commit()

    def open_gzip_index(self) -> Optional[SQLiteBlobsFile]:
        connection = self.get_connection()
        tables = get_sqlite_tables(connection)
        if 'gzipindex' in tables or 'gzipindexes' in tables:
            table = 'gzipindexes' if 'gzipindexes' in tables else 'gzipindex'
            return SQLiteBlobsFile(connection, table, 'data', buffer_size=SQLiteIndex._MAX_BLOB_SIZE)
        return None



# pylint: disable=wrong-import-position
# pylint: disable=protected-access

import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from ratarmountcore.utils import (  # noqa: E402
    ALPHA,
    DIGITS,
    HEX,
    LRUCache,
    Prefetcher,
    ceil_div,
    format_number,
    is_latin_alpha,
    is_latin_digit,
    is_latin_hex_alpha,
    is_random,
)


def test_ceil_div():
    assert ceil_div(0, 1) == 0
    assert ceil_div(0, 1000) == 0
    assert ceil_div(0, -1000) == 0

    assert ceil_div(1, 1000) == 1
    assert ceil_div(1, 1) == 1

    assert ceil_div(0, 2) == 0
    assert ceil_div(1, 2) == 1
    assert ceil_div(2, 2) == 1
    assert ceil_div(3, 2) == 2
    assert ceil_div(4, 2) == 2
    assert ceil_div(5, 2) == 3


def test_prefetcher():
    memorySize = 4
    prefetcher = Prefetcher(memorySize)

    # Without any knowledge, it can't give any prefetch suggestion
    assert not prefetcher.prefetch(0)
    assert not prefetcher.prefetch(1)
    assert not prefetcher.prefetch(2)
    assert not prefetcher.prefetch(2)

    assert prefetcher.fetch(32) is None

    # With a single known sample, we can not interpolate anything, so simply return a single example, the next one
    assert not list(prefetcher.prefetch(0))
    assert list(prefetcher.prefetch(1)) == [33]
    assert list(prefetcher.prefetch(2)) == [33]
    assert list(prefetcher.prefetch(3)) == [33]

    assert prefetcher.fetch(33) is None

    assert not list(prefetcher.prefetch(0))
    assert list(prefetcher.prefetch(1)) == [34]
    assert list(prefetcher.prefetch(4)) == [34, 35]

    assert prefetcher.fetch(34) is None
    assert prefetcher.fetch(35) is None

    # When the last fetched cache is fully filled with a monotonic increasing series,
    # then always prefetch the full requested amount
    for n in range(32):
        assert list(prefetcher.prefetch(n)) == list(range(36, 36 + n))

    assert prefetcher.fetch(36) is None
    assert prefetcher.fetch(37) is None

    # Same as above but the cache is limited, so fetching anything more
    # does not change the length of the returned result further
    for n in range(32):
        assert list(prefetcher.prefetch(n)) == list(range(38, 38 + n))

    prefetcher.fetch(0)
    assert list(prefetcher.prefetch(8)) == [1]


def test_lru_cache():
    cache = LRUCache(3)

    assert not cache
    assert len(cache) == 0

    assert cache.get(5) is None

    for n, key, value in zip(range(cache.size), range(5, 5 + cache.size), range(15, 15 + cache.size)):
        # Repeated identical insertions are stable
        for _ in range(10):
            cache[key] = value
            assert len(cache) == n + 1
            assert cache.get(key) == value
            assert cache.get(key) == value

    # Inserting a key when the maximum size has been reached, will evict the least recently used key
    assert len(cache) == cache.size
    cache[32] = 132
    assert len(cache) == cache.size
    # Note that even these get tests will change the least recently used status of the keys
    assert cache.get(32) == 132
    assert cache.get(5) is None
    assert cache.get(6) == 16
    assert cache.get(7) == 17
    assert cache[7] == 17

    cache[33] = 133
    assert 32 not in cache
    assert cache[33] == 133

    assert sorted(cache.keys()) == [6, 7, 33]
    del cache[33]
    assert len(cache) == 2
    assert 33 not in cache
    assert sorted(cache.keys()) == [6, 7]

    cache.clear()
    assert not cache
    assert len(cache) == 0
    assert 7 not in cache
    assert not cache.keys()


def test_is_latin_alpha():
    assert ALPHA == 'abcdefghijklmnopqrstuvwxyz'
    assert is_latin_alpha(ALPHA)

    assert is_latin_alpha("a")
    assert is_latin_alpha("ab")
    assert is_latin_alpha("z")
    assert not is_latin_alpha("0")
    assert not is_latin_alpha("")

    # Python's isalpha is too Unicode. I only want to match the usual
    # ASCII Latin alphabet not mathematical or Greek symbols!
    # https://util.unicode.org/UnicodeJsps/confusables.jsp
    # https://www.unicode.org/Public/security/revision-03/confusablesSummary.txt
    assert "α".isalpha()
    assert not is_latin_alpha("α")

    assert "𝒻".isalpha()
    assert not is_latin_alpha("𝒻")


def test_is_latin_digit():
    assert DIGITS == '0123456789'
    assert is_latin_digit(DIGITS)

    assert is_latin_digit("0")
    assert is_latin_digit("10")
    assert is_latin_digit("9")
    assert not is_latin_digit("a")
    assert not is_latin_digit("")

    # Python's isdigit is too Unicode. I only want to match the usual
    # ASCII Latin digits not decimal digits from other languages!
    # https://www.compart.com/en/unicode/category/Nd
    assert "௫".isdigit()
    assert not is_latin_digit("௫")

    assert "߂".isdigit()
    assert not is_latin_digit("߂")


def test_is_latin_hex_alpha():
    assert HEX == '0123456789abcdef'
    assert is_latin_hex_alpha(HEX)

    assert is_latin_hex_alpha("0")
    assert is_latin_hex_alpha("10")
    assert is_latin_hex_alpha("9")
    assert is_latin_hex_alpha("a")
    assert is_latin_hex_alpha("f")
    assert not is_latin_hex_alpha("g")
    assert not is_latin_hex_alpha("")

    # Python's isdigit is too Unicode. I only want to match the usual
    # ASCII Latin digits not decimal digits from other languages!
    # https://www.compart.com/en/unicode/category/Nd
    assert "௫".isdigit()
    assert not is_latin_hex_alpha("௫")

    assert "߂".isdigit()
    assert not is_latin_hex_alpha("߂")


def test_format_number():
    assert format_number(0, ALPHA) == 'a'
    assert format_number(1, ALPHA) == 'b'
    assert format_number(25, ALPHA) == 'z'
    assert format_number(26, ALPHA) == 'ba'
    assert format_number(26 * 26, ALPHA) == 'baa'
    assert format_number(26 * 26 + 3, ALPHA) == 'bad'

    assert format_number(0, ALPHA, 3) == 'aaa'
    assert format_number(1, ALPHA, 3) == 'aab'
    assert format_number(25, ALPHA, 3) == 'aaz'
    assert format_number(26, ALPHA, 3) == 'aba'
    assert format_number(26 * 26, ALPHA, 3) == 'baa'
    assert format_number(26 * 26 + 3, ALPHA, 3) == 'bad'

    assert format_number(0, HEX) == '0'
    assert format_number(1, HEX) == '1'
    assert format_number(9, HEX) == '9'
    assert format_number(10, HEX) == 'a'
    assert format_number(15, HEX) == 'f'
    assert format_number(16, HEX) == '10'
    assert format_number(16 * 16, HEX) == '100'
    assert format_number(16 * 16 + 3, HEX) == '103'

    assert format_number(0, HEX, 3) == '000'
    assert format_number(1, HEX, 3) == '001'
    assert format_number(9, HEX, 3) == '009'
    assert format_number(10, HEX, 3) == '00a'
    assert format_number(15, HEX, 3) == '00f'
    assert format_number(16, HEX, 3) == '010'
    assert format_number(16 * 16, HEX, 3) == '100'
    assert format_number(16 * 16 + 3, HEX, 3) == '103'

    assert format_number(357641610, DIGITS) == '357641610'


def test_is_random():
    data = os.urandom(1 << 20)
    for size in [1024, 1280, 2048, 3333, 4096, len(data) // 100, len(data) // 10, len(data) // 2, len(data)]:
        assert is_random(data[:size])

    data = bytes(i for i in range(256))
    assert not is_random(data)
    assert not is_random(data * 4)
    assert not is_random(data * 100)
    # Too few bytes to determine randomness! Also no repeated letter! I guess the random test would have to check
    # specifically for some bit distributions because these are mostly ASCII characters, i.e., the 7-th bit is not
    # set. Or maybe some clustering analysis because the bytes are clustered in a very close range of bytes.
    # But oh well. Simply require >= 256 input bytes.
    # assert is_random(b'SQLite format 3\x00')
    # assert is_random(b'abcde')
    # assert is_random(b'abcdefghijklmnopqrstuvxyz')
    # assert is_random(b'aaaaa')
